package precision.model;

import java.security.InvalidParameterException;

/**
 * The double value what is prepared for "equal" and "compare" operations with respect to the Model of precise calculations and Accuracy variable value.
 * 
 * The instances of this type doesn't track runtime changes of the Accuracy variable.
 */
public final class PreciseDouble implements Comparable<PreciseDouble>
{
	private final byte sign;
	
	private final long wholePart;
	
	private final long decimalPart;
	
	private int hashCode;
	
	private final double rawDouble;
	
	public PreciseDouble(double value, IPrecisionConfiguration precisionConfiguration)
	{
		if (precisionConfiguration == null)
			throw new InvalidParameterException();
		
		this.rawDouble = value;
		
		double sign = Math.signum(value);

		long wholePart = (long)(sign * value);
		
		double reminder = sign * value - wholePart;
		long decimalPart = (long)Math.rint(precisionConfiguration.iEpsilon() * reminder);
		
		if (decimalPart == precisionConfiguration.iEpsilon())
		{
			this.wholePart = wholePart + 1;
			this.decimalPart = 0;
		} else
		{
			this.wholePart = wholePart;
			this.decimalPart = decimalPart;
		}

		if (this.wholePart == 0 && this.decimalPart == 0)
			this.sign = 0;
		else
			this.sign = (byte)sign;
	}
	
	/**
	 * Return 0 if value is  (or under precision interval), 1 if value is greater than 0, -1 if value is less than 0. 
	 */
	public final byte sign()
	{
		return this.sign;
	}
	
	/**
	 * The whole part of the number without its sign.
	 * @return
	 */
	public final long wholePart()
	{
		return this.wholePart;
	}
	
	/**
	 * Decimal part of the number without its sign.
	 * @return
	 */
	public final long decimalPart()
	{
		return this.decimalPart;
	}
	
	public final double rawDouble()
	{
		return this.rawDouble;
	}
	
	@Override
	public final boolean equals(Object other) 
	{
		if (other == null || this.getClass() != other.getClass())
			return false;
		if (this == other)
			return true;
		PreciseDouble otherAccurateDouble = (PreciseDouble) other;
		return this.sign == otherAccurateDouble.sign && this.wholePart == otherAccurateDouble.wholePart && this.decimalPart == otherAccurateDouble.decimalPart;
	}
	
	@Override
	public final int hashCode() 
	{
		if (this.hashCode == 0)
		{
			int hc = 41;
			hc = (hc * 43)  ^ this.sign;
			hc = (hc * 43) ^ Long.valueOf(this.wholePart).hashCode();
			hc = (hc * 43) ^ Long.valueOf(this.decimalPart).hashCode();
			this.hashCode = hc;
		}
		return this.hashCode;
	}

	@Override
	public final int compareTo(PreciseDouble o) 
	{
		if (this.wholePart == o.wholePart && this.wholePart == 0)
		{
			if (this.sign == o.sign)
				return 0;
			else if (this.sign < o.sign)
				return -1;
			else
				return 1;				
		}
		else if (this.wholePart > o.wholePart)
		{
			return 1;
		} 
		else if (this.wholePart < o.wholePart)
		{
			return -1;
		} 
		else
		{
			if (this.decimalPart > o.decimalPart)
			{
				return 1;
			} 
			else if (this.decimalPart < o.decimalPart)				
			{
				return -1;
			} 
			else
			{
				return 0;
			}
		}		
	}
}