package spatial.algorithms;

import precision.model.IPrecisionConfiguration;
import spatial.algorithms.bestFittingPlane.BestFittingPlaneAlgorithm;
import spatial.algorithms.csTransform.CsTransformationAlgorithm;
import spatial.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationAlgorithm;
import spatial.algorithms.edgeToFaceLocation.EdgeToFaceLocationAlgorithm;
import spatial.algorithms.edgeToPolygonLocation.EdgeToPolygonLocationAlgorithm;
import spatial.algorithms.faceArea.FaceAreaAlgorithm;
import spatial.algorithms.pointToEdgeDistance.PointToEdgeDistanceAlgorithm;
import spatial.algorithms.pointToFaceLocation.PointToFaceLocationAlgorithm;
import spatial.algorithms.pointToLineLocation.PointToLineLocationAlgorithm;
import spatial.algorithms.pointToPlaneDistance.PointToPlaneDistanceAlgorithm;
import spatial.algorithms.pointToPolygonDistance.PointToPolygonDistanceAlgorithm;
import spatial.algorithms.pointToPolygonLocation.PointToPolygonLocationAlgorithm;
import spatial.algorithms.polygon2dSelfIntersection.Polygon2dSelfIntersectionAlgorithm;
import spatial.algorithms.polygonArea.PolygonAreaAlgorithm;
import spatial.algorithms.polygonToPolygonLocation.PolygonToPolygonLocationAlgorithm;
import spatial.geometry.Edge2d;
import spatial.geometry.Edge3d;
import spatial.geometry.Face2d;
import spatial.geometry.Face3d;
import spatial.geometry.Line3d;
import spatial.geometry.Plane3d;
import spatial.geometry.Polygon2d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.ICommonGeometryAlgorithms;
import spatial.model.algorithms.bestFittingPlane.IBestFittingPlaneResult;
import spatial.model.algorithms.csTransformation.ICsTransformationResult;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;
import spatial.model.algorithms.edgeToFaceLocation.IEdgeToFaceLocationResult;
import spatial.model.algorithms.edgeToPolygonLocation.IEdgeToPolygonLocationResult;
import spatial.model.algorithms.faceArea.IFaceAreaResult;
import spatial.model.algorithms.pointToEdgeDistance.IPointToEdgeDistanceResult;
import spatial.model.algorithms.pointToFaceLocation.IPointToFaceLocationResult;
import spatial.model.algorithms.pointToLineLocation.IPointToLineLocationResult;
import spatial.model.algorithms.pointToPlaneDistance.IPointToPlaneDistanceResult;
import spatial.model.algorithms.pointToPolygonDistance.IPointToPolygonDistanceResult;
import spatial.model.algorithms.pointToPolygonLocation.IPointToPolygonLocationResult;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionResult;
import spatial.model.algorithms.polygonArea.IPolygonAreaResult;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationResult;

public class CommonGeometryAlgorithms implements ICommonGeometryAlgorithms
{
	private final IPrecisionConfiguration precisionConfiguration;

	public CommonGeometryAlgorithms(IPrecisionConfiguration precisionConfiguration)
	{
		if (precisionConfiguration == null)
			throw new IllegalArgumentException("Precision configuration is not defined.");
		this.precisionConfiguration = precisionConfiguration;
	}

	@Override
	public IBestFittingPlaneResult bestFittingPlane(Vector3d[] points)
	{
		return (new BestFittingPlaneAlgorithm(this.precisionConfiguration)).calculate(points);
	}

	@Override
	public ICsTransformationResult csTransformation(Vector3d normalPlaneOrt, Vector3d inPlaneOrt,
			Vector3d localCSOrigin, Vector3d[] globalCsPoints)
	{
		return (new CsTransformationAlgorithm(this.precisionConfiguration)).calculate(normalPlaneOrt, inPlaneOrt, localCSOrigin, globalCsPoints);
	}

	@Override
	public IEdgeToEdgeLocationResult edgeToEdgeLocation(Edge2d edge1, Edge2d edge2)
	{
		return (new EdgeToEdgeLocationAlgorithm(this.precisionConfiguration)).calculate(edge1, edge2);
	}

	@Override
	public IEdgeToFaceLocationResult edgeToFaceLocation(Vector3d planeNormalUnitVector, Face3d face, Edge3d testEdge)
	{
		return (new EdgeToFaceLocationAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, face, testEdge);
	}

	@Override
	public IEdgeToPolygonLocationResult edgeToPolygonLocation(Vector3d planeNormalUnitVector, Polygon3d polygon,
			Edge3d testEdge)
	{
		return (new EdgeToPolygonLocationAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, polygon, testEdge);
	}

	@Override
	public IFaceAreaResult faceArea(Face2d face)
	{
		return (new FaceAreaAlgorithm(this.precisionConfiguration)).calculate(face);
	}

	@Override
	public IPointToEdgeDistanceResult pointToEdgeDistance(Edge3d edge, Vector3d point)
	{
		return (new PointToEdgeDistanceAlgorithm(this.precisionConfiguration)).calculate(edge, point);
	}

	@Override
	public IPointToFaceLocationResult pointToFaceLocation(Vector3d planeNormalUnitVector, Face3d face, Vector3d testPoint)
	{
		return (new PointToFaceLocationAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, face, testPoint);
	}

	@Override
	public IPointToLineLocationResult pointToLineLocation(Vector3d planeNormalUnitVector, Line3d line, Vector3d testPoint)
	{
		return (new PointToLineLocationAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, line, testPoint);
	}

	@Override
	public IPointToPlaneDistanceResult pointToPlaneDistance(Plane3d plane, Vector3d testPoint)
	{
		return (new PointToPlaneDistanceAlgorithm(this.precisionConfiguration)).calculate(plane, testPoint);
	}

	@Override
	public IPointToPolygonDistanceResult pointToPolygonDistanceResult(Vector3d planeNormalUnitVector, Polygon3d polygon,
			Vector3d testPoint)
	{
		return (new PointToPolygonDistanceAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, polygon, testPoint);
	}

	@Override
	public IPointToPolygonLocationResult pointToPolygonLocation(Vector3d planeNormalUnitVector, Polygon3d polygon, Vector3d testPoint)
	{
		return (new PointToPolygonLocationAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, polygon, testPoint);
	}

	@Override
	public IPolygon2dSelfIntersectionResult polygon2dSelfIntersection(Vector3d planeNormalUnitVector, Polygon3d polygon)
	{
		return (new Polygon2dSelfIntersectionAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, polygon);
	}

	@Override
	public IPolygonAreaResult polygonArea(Polygon2d polygon)
	{
		return (new PolygonAreaAlgorithm(this.precisionConfiguration)).calculate(polygon);
	}

	@Override
	public IPolygonToPolygonLocationResult polygonToPolygonLocation(Vector3d planeNormalUnitVector, Polygon3d polygon1,
			Polygon3d polygon2)
	{
		return (new PolygonToPolygonLocationAlgorithm(this.precisionConfiguration)).calculate(planeNormalUnitVector, polygon1, polygon2);
	}

}
