package spatial.algorithms.bestFittingPlane;

import org.ejml.data.DMatrixRMaj;
import org.ejml.dense.row.factory.DecompositionFactory_DDRM;
import org.ejml.interfaces.decomposition.SingularValueDecomposition;

import precision.model.IPrecisionConfiguration;
import spatial.algorithms.AlgorithmBase;
import spatial.geometry.Plane3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.bestFittingPlane.IBestFittingPlaneAlgorithm;
import spatial.model.algorithms.bestFittingPlane.IBestFittingPlaneResult;

public final class BestFittingPlaneAlgorithm extends AlgorithmBase<IBestFittingPlaneResult> implements IBestFittingPlaneAlgorithm
{
	//start: algorithm input arguments

	private Vector3d[] points;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private Plane3d bestFittingPlane;
	
	@SuppressWarnings("unused")
	private double minimalSingularValue;
	
	private double[] pointsToPlaneDistances;
	
	//end: algorithm result arguments

	public BestFittingPlaneAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public final IBestFittingPlaneResult calculate(Vector3d[] points)
	{
		if (points == null)
			throw new IllegalArgumentException("The Points are not defined.");
		
		this.points = points;
		
		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
		if (points.length < 3)
		{
			this.addError("The number of defined points is less than 3.");
			return;
		}
		
		//calculate centroid
		double x = 0, y = 0, z = 0;
		for (int i = 0; i < this.points.length; i++)
		{
			x += this.points[i].x();
			y += this.points[i].y();
			z += this.points[i].z();
		}
		Vector3d centroid = new Vector3d(x / this.points.length, y / this.points.length, z / this.points.length, this.precisionConfiguration);

		//create matrix from the points moved to the centre of CS
		double[] dataPoints = new double[this.points.length * 3];
		for (int i = 0; i < this.points.length; i++)
		{
			dataPoints[i * 3] = this.points[i].x() - centroid.x();
			dataPoints[i * 3 + 1] = this.points[i].y() - centroid.y();
			dataPoints[i * 3 + 2] = this.points[i].z() - centroid.z();
		}
		DMatrixRMaj mx = new DMatrixRMaj(this.points.length, 3, true, dataPoints);
		
		//svd decomposition
		SingularValueDecomposition<DMatrixRMaj> svd = DecompositionFactory_DDRM.svd(false, true, true);
		boolean decomposed = svd.decompose(mx);
		
		if(decomposed)
		{
			DMatrixRMaj wmx = svd.getW(null);
			DMatrixRMaj vmx = svd.getV(null, false);
			
			//determine the index of the minimal singular value
			int minIx = 0;
			for (int i = 1; i < wmx.numCols; i++)
			{
				if (wmx.get(i, i) < wmx.get(minIx, minIx))
					minIx = i;
			}

			this.minimalSingularValue = wmx.get(minIx, minIx);
			
			Vector3d planeNormalUnitVector = new Vector3d(
					vmx.get(0, minIx),
					vmx.get(1, minIx),
					vmx.get(2, minIx),
					this.precisionConfiguration);
			
			this.bestFittingPlane = new Plane3d(planeNormalUnitVector, centroid);
			
			//calculate distances for every source point from the plane
			this.pointsToPlaneDistances = new double[this.points.length];
			for (int i = 0; i < this.points.length; i++)
			{
				Vector3d x0ToSourcePointVector = this.points[i].substract(this.bestFittingPlane.pointOnThePlane(), this.precisionConfiguration);
				double distance = this.bestFittingPlane.normalUnitVector().dotProduct(x0ToSourcePointVector);
				this.pointsToPlaneDistances[i] = distance;
			}
		} 
		else
		{
			this.addError("Unknown problem with svd decomposition.");
		}
		
	}

	@Override
	protected IBestFittingPlaneResult constructResult()
	{
		return new BestFittingPlaneResult(this.bestFittingPlane, this.pointsToPlaneDistances);
	}

	@Override
	protected IBestFittingPlaneResult constructErrorsResult()
	{
		return new BestFittingPlaneResult(this.errors());
	}

}
