package spatial.algorithms.edgeToEdgeLocation;

import precision.model.IPrecisionConfiguration;
import spatial.algorithms.AlgorithmBase;
import spatial.geometry.Edge2d;
import spatial.geometry.Vector2d;
import spatial.model.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationKind;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationAlgorithm;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;

public final class EdgeToEdgeLocationAlgorithm extends AlgorithmBase<IEdgeToEdgeLocationResult> implements IEdgeToEdgeLocationAlgorithm
{
	//start: algorithm input arguments
	
	private Edge2d abEdge;
	
	private Edge2d cdEdge;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private EdgeToEdgeLocationKind locationCode;
	
	private double r0;
	
	private double t0;
	
	private double rc;
	
	private double rd;
	
	private Vector2d intersectionPoint;

	//end: algorithm result arguments
	
	public EdgeToEdgeLocationAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public final IEdgeToEdgeLocationResult calculate(Edge2d edge1, Edge2d edge2)
	{
		if (edge1 == null)
			throw new IllegalArgumentException("The Edge 1 is not defined.");
		if (edge2 == null)
			throw new IllegalArgumentException("The Edge 2 is not defined.");
		
		this.abEdge = edge1;
		this.cdEdge = edge2;
		
		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
		double a1 = this.abEdge.vectorA().x();
		double a2 = this.abEdge.vectorA().y();
		double b1 = this.abEdge.vectorB().x();
		double b2 = this.abEdge.vectorB().y();
		
		double c1 = this.cdEdge.vectorA().x();
		double c2 = this.cdEdge.vectorA().y();
		double d1 = this.cdEdge.vectorB().x();
		double d2 = this.cdEdge.vectorB().y();
		
		double det = (b1 - a1) * (c2 - d2) - (b2 - a2) * (c1 - d1);
		
		if (this.precisionConfiguration.compareWithAccuracy(0, det) != 0)
		{
			this.r0 = ((c1 - a1) * (c2 - d2) - (c2 - a2) * (c1 - d1)) / det;
			this.t0 = ((c2 - a2) * (b1 - a1) - (c1 - a1) * (b2 - a2)) / det;

			double s1 = (1 - this.r0) * a1 + this.r0 * b1;
			double s2 = (1 - this.r0) * a2 + this.r0 * b2;
			
			this.intersectionPoint = new Vector2d(s1, s2, this.precisionConfiguration);
			
			this.locationCode = EdgeToEdgeLocationKind.NotParallel;
		} 
		else
		{
			double area = (b1 - a1) * (c2 -a2) - (b2 - a2) * (c1 - a1);
			
			if (this.precisionConfiguration.compareWithAccuracy(0, area) != 0)
			{
				this.locationCode = EdgeToEdgeLocationKind.Parallel;
			}
			else
			{
				double divider = (b1 - a1) * (b1 - a1) + (b2 - a2) * (b2 - a2);				
				this.rc = ((c1 - a1) * (b1 - a1) + (c2 - a2) * (b2 - a2)) / divider;
				this.rd = ((d1 - a1) * (b1 - a1) + (d2 - a2) * (b2 - a2)) / divider;
				
				if (this.precisionConfiguration.compareWithAccuracy(this.rc, 0) == -1 && this.precisionConfiguration.compareWithAccuracy(this.rd, 0) == -1 ||
					this.precisionConfiguration.compareWithAccuracy(this.rc, 1) == 1 && this.precisionConfiguration.compareWithAccuracy(this.rd, 1) == 1)
				{
					this.locationCode = EdgeToEdgeLocationKind.Collinear;
				}
				else
				{
					if (this.precisionConfiguration.compareWithAccuracy(this.rc, 0) == -1 && this.precisionConfiguration.compareWithAccuracy(this.rd, 0) == 0 ||
						this.precisionConfiguration.compareWithAccuracy(this.rd, 0) == -1 && this.precisionConfiguration.compareWithAccuracy(this.rc, 0) == 0 ||
						this.precisionConfiguration.compareWithAccuracy(this.rc, 1) == 1 && this.precisionConfiguration.compareWithAccuracy(this.rd, 1) == 0 ||
						this.precisionConfiguration.compareWithAccuracy(this.rd, 1) == 1 && this.precisionConfiguration.compareWithAccuracy(this.rc, 1) == 0)
					{
						this.locationCode = EdgeToEdgeLocationKind.OverlapAtNode;
					} 
					else
					{
						this.locationCode = EdgeToEdgeLocationKind.Overlap;
					}
				}
			}
		}
	}

	@Override
	protected IEdgeToEdgeLocationResult constructResult()
	{
		return new EdgeToEdgeLocationResult(this.locationCode, this.r0, this.t0, this.rc, this.rd, this.intersectionPoint, this.precisionConfiguration);
	}

	@Override
	protected IEdgeToEdgeLocationResult constructErrorsResult()
	{
		return new EdgeToEdgeLocationResult(this.errors());
	}

}
