package spatial.algorithms.edgeToFaceLocation;

import precision.model.IPrecisionConfiguration;
import spatial.algorithms.AlgorithmBase;
import spatial.algorithms.edgeToPolygonLocation.EdgeToPolygonLocationAlgorithm;
import spatial.geometry.ClockWiseType;
import spatial.geometry.Edge3d;
import spatial.geometry.Face3d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.edgeToFaceLocation.EdgeToFaceLocationKind;
import spatial.model.algorithms.edgeToFaceLocation.IEdgeToFaceLocationAlgorithm;
import spatial.model.algorithms.edgeToFaceLocation.IEdgeToFaceLocationResult;
import spatial.model.algorithms.edgeToPolygonLocation.EdgeToPolygonLocationKind;
import spatial.model.algorithms.edgeToPolygonLocation.IEdgeToPolygonLocationResult;

public final class EdgeToFaceLocationAlgorithm extends AlgorithmBase<IEdgeToFaceLocationResult> implements IEdgeToFaceLocationAlgorithm
{

	//start: algorithm input arguments

	private Vector3d planeNormalUnitVector;
	
	private Face3d faceGeometry;
	
	private Edge3d testEdge;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private EdgeToFaceLocationKind edgeToFaceLocationKind;
	
	//end: algorithm result arguments

	public EdgeToFaceLocationAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public IEdgeToFaceLocationResult calculate(Vector3d planeNormalUnitVector, Face3d face, Edge3d testEdge)
	{
		if (planeNormalUnitVector == null)
			throw new IllegalArgumentException("Plane normal unit vector is not defined");
		if (face == null)
			throw new IllegalArgumentException("The face is not defined");
		if (testEdge == null)
			throw new IllegalArgumentException("The test edge is not defined");
		
		this.planeNormalUnitVector = planeNormalUnitVector;
		this.faceGeometry = face;
		this.testEdge = testEdge;

		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
		Polygon3d outerPolygon = null;
		Polygon3d[] innerPolygons = new Polygon3d[this.faceGeometry.polygons().length - 1];
		int ipIndex = 0;
		
		for (int i = 0; i < this.faceGeometry.polygons().length; i++)
		{
			Polygon3d p = this.faceGeometry.polygons()[i];
			
			ClockWiseType cwt = p.clockWiseType(this.planeNormalUnitVector, this.precisionConfiguration);
			
			if (cwt == ClockWiseType.CounterClockWise)
			{
				if (outerPolygon == null)
				{
					outerPolygon = p;
				}
				else
				{
					this.addError("Faces with more than 1 ccw-polygons are not supported.");
					return;
				}
			}
			else if (cwt == ClockWiseType.ClockWise)
			{
				if (ipIndex < innerPolygons.length)
				{
					innerPolygons[ipIndex] = p;
				} 
				else
				{
					this.addError("Faces with more than no ccw-polygons are not supported.");
					return;
				}
			}
			else
			{
				this.addError("Face isn't plane.");
				return;
			}
		}
		
		//test outer polygon
		
		IEdgeToPolygonLocationResult outerPolygonLocationResult = (new EdgeToPolygonLocationAlgorithm(this.precisionConfiguration))
				.calculate(this.planeNormalUnitVector, outerPolygon, this.testEdge);
		
		if (outerPolygonLocationResult.hasErrors())
		{
			this.addErrors(outerPolygonLocationResult.errors());
			return;
		}
		
		if (outerPolygonLocationResult.locationKind() == EdgeToPolygonLocationKind.Error)
		{
			this.edgeToFaceLocationKind = EdgeToFaceLocationKind.Error;
			return;
		}
		else if (outerPolygonLocationResult.locationKind() == EdgeToPolygonLocationKind.Exterior)
		{
			this.edgeToFaceLocationKind = EdgeToFaceLocationKind.Exterior;
			return;
		}
		
		//test inner polygons
		for (int i = 0; i < innerPolygons.length; i++)
		{
			IEdgeToPolygonLocationResult result = (new EdgeToPolygonLocationAlgorithm(this.precisionConfiguration))
					.calculate(this.planeNormalUnitVector, innerPolygons[i], this.testEdge);
			
			if (result.hasErrors())
			{
				this.addErrors(result.errors());
				return;
			}
			
			if (result.locationKind() == EdgeToPolygonLocationKind.Error)
			{
				this.edgeToFaceLocationKind = EdgeToFaceLocationKind.Error;
				return;
			}
			else if (result.locationKind() == EdgeToPolygonLocationKind.Exterior)
			{
				this.edgeToFaceLocationKind = EdgeToFaceLocationKind.Exterior;
				return;
			}
		}
		
		this.edgeToFaceLocationKind = EdgeToFaceLocationKind.Interior;
	}

	@Override
	protected IEdgeToFaceLocationResult constructResult()
	{
		return new EdgeToFaceLocationResult(this.edgeToFaceLocationKind);
	}

	@Override
	protected IEdgeToFaceLocationResult constructErrorsResult()
	{
		return new EdgeToFaceLocationResult(this.errors());
	}

	
}
