package spatial.algorithms.edgeToPolygonLocation;

import precision.model.IPrecisionConfiguration;
import spatial.algorithms.AlgorithmBase;
import spatial.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationAlgorithm;
import spatial.algorithms.pointToPolygonLocation.PointToPolygonLocationAlgorithm;
import spatial.geometry.CoordinateSystem3d;
import spatial.geometry.Edge2d;
import spatial.geometry.Edge3d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector2d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationKind;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationAlgorithm;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;
import spatial.model.algorithms.edgeToPolygonLocation.EdgeToPolygonLocationKind;
import spatial.model.algorithms.edgeToPolygonLocation.IEdgeToPolygonLocationAlgorithm;
import spatial.model.algorithms.edgeToPolygonLocation.IEdgeToPolygonLocationResult;
import spatial.model.algorithms.pointToPolygonLocation.IPointToPolygonLocationAlgorithm;
import spatial.model.algorithms.pointToPolygonLocation.IPointToPolygonLocationResult;
import spatial.model.algorithms.pointToPolygonLocation.PointToPolygonLocationKind;

public final class EdgeToPolygonLocationAlgorithm extends AlgorithmBase<IEdgeToPolygonLocationResult> implements IEdgeToPolygonLocationAlgorithm
{
	//start: algorithm input arguments

	private Vector3d planeNormalUnitVector;
	
	private Polygon3d polygon;
	
	private Edge3d testEdge;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private EdgeToPolygonLocationKind edgeToPolygonLocationKind;
	
	//end: algorithm result arguments
	
	public EdgeToPolygonLocationAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public IEdgeToPolygonLocationResult calculate(Vector3d planeNormalUnitVector, Polygon3d polygon, Edge3d testEdge)
	{
		if (planeNormalUnitVector == null)
			throw new IllegalArgumentException("The Plane Normal Unit Vector is not defined.");
		if (polygon == null)
			throw new IllegalArgumentException("The Polygon is not defined.");
		if (testEdge == null)
			throw new IllegalArgumentException("The Test Edge is not defined.");
		
		this.planeNormalUnitVector = planeNormalUnitVector;
		this.polygon = polygon;
		this.testEdge = testEdge;
		
		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
		//move input to polygon local coordinate system
		CoordinateSystem3d localCoordinateSystem = this.createLocalCoordinateSystem();
		
		Polygon3d polygonLocal3d = this.transformInputPolygonToLocalCoordinateSystem(localCoordinateSystem);
		Edge3d testEdgeLocal3d = new Edge3d( localCoordinateSystem.toLocal(this.testEdge.vectorA(), this.precisionConfiguration),
				localCoordinateSystem.toLocal(this.testEdge.vectorB(), this.precisionConfiguration));

		//test if polygon is plane and test edge is in the plane too
		if (!this.isPolygonPlane(polygonLocal3d))
		{
			this.addError("Polygon A isn't plane.");
			return;
		}
		if (this.precisionConfiguration.compareWithAccuracy(testEdgeLocal3d.vectorA().z(), 0) != 0 ||
				this.precisionConfiguration.compareWithAccuracy(testEdgeLocal3d.vectorB().z(), 0) != 0)
		{
			this.addError("The Test Edge is not in the polygon plane");
			return;
		}
		
		Edge2d testEdgeLocal2d = this.edge3dTo2d(testEdgeLocal3d);
		
		//test each polygon edge for intersection with test edge
		for	(int i = 0; i < polygonLocal3d.vertices().length; i++)
		{
			Edge2d pEdge2d = this.edge3dTo2d(polygonLocal3d.getEdge(i));
			
			IEdgeToEdgeLocationAlgorithm edgeToEdgeLocationAlgorithm = new EdgeToEdgeLocationAlgorithm(this.precisionConfiguration);
			IEdgeToEdgeLocationResult result = edgeToEdgeLocationAlgorithm.calculate(pEdge2d, testEdgeLocal2d);
			
			if (result.locationCode() != EdgeToEdgeLocationKind.Collinear && result.locationCode() != EdgeToEdgeLocationKind.OverlapAtNode && result.locationCode() != EdgeToEdgeLocationKind.Parallel)
			{
				if (result.locationCode() == EdgeToEdgeLocationKind.Overlap)
				{
					this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Error;
					return;
				} 
				else if (result.locationCode() == EdgeToEdgeLocationKind.NotParallel)
				{
					if (this.edgesIntersectInside(result.r0(), result.t0()))
					{
						this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Error;
						return;
					}
				} 
				else
				{
					this.addError("Not expected location code for test edge and polygon edge intersection");
					return;
				}
			}
		}

		//test location for test edge nodes
		PointToPolygonLocationKind edgeALocation = this.getPointToPolygonLocation(this.testEdge.vectorA());
		PointToPolygonLocationKind edgeBLocation = this.getPointToPolygonLocation(this.testEdge.vectorB());
		
		if (edgeALocation == PointToPolygonLocationKind.NotSet || edgeBLocation == PointToPolygonLocationKind.NotSet)
		{
			return;
		}
		else if (edgeALocation == PointToPolygonLocationKind.OnBoundary && edgeBLocation == PointToPolygonLocationKind.OnBoundary)
		{
			Vector3d midPoint = this.testEdge.vectorA().add(this.testEdge.edgeVector(this.precisionConfiguration).mul(0.5, this.precisionConfiguration), precisionConfiguration);
			
			PointToPolygonLocationKind midPointLocation = this.getPointToPolygonLocation(midPoint);
			
			if (midPointLocation == PointToPolygonLocationKind.NotSet)
			{
				return;
			}
			else if (midPointLocation == PointToPolygonLocationKind.Inside)
			{
				this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Interior;
			}
			else if (midPointLocation == PointToPolygonLocationKind.Outside)
			{
				this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Exterior;
			}
			else
			{
				this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Error;
			}
		}
		else if ((edgeALocation == PointToPolygonLocationKind.Inside || edgeALocation == PointToPolygonLocationKind.OnBoundary) &&
				(edgeBLocation == PointToPolygonLocationKind.Inside || edgeBLocation == PointToPolygonLocationKind.OnBoundary))
		{
			this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Interior;
		} 
		else if ((edgeALocation == PointToPolygonLocationKind.Outside || edgeALocation == PointToPolygonLocationKind.OnBoundary) &&
				(edgeBLocation == PointToPolygonLocationKind.Outside || edgeBLocation == PointToPolygonLocationKind.OnBoundary))
		{
			this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Exterior;
		} 
		else
		{
			this.edgeToPolygonLocationKind = EdgeToPolygonLocationKind.Error;
		}
		
	}

	@Override
	protected IEdgeToPolygonLocationResult constructResult()
	{
		return new EdgeToPolygonLocationResult(this.edgeToPolygonLocationKind);
	}

	@Override
	protected IEdgeToPolygonLocationResult constructErrorsResult()
	{
		return new EdgeToPolygonLocationResult(this.errors());
	}

	private PointToPolygonLocationKind getPointToPolygonLocation(Vector3d point)
	{
		IPointToPolygonLocationAlgorithm pointToPolygonLocationAlgorithm = new PointToPolygonLocationAlgorithm(this.precisionConfiguration);
		IPointToPolygonLocationResult result = pointToPolygonLocationAlgorithm.calculate(this.planeNormalUnitVector, this.polygon, point);
		if (result.hasErrors())
		{
			this.addErrors(result.errors());
			return PointToPolygonLocationKind.NotSet;
		}
		else
		{
			try
			{
				return result.locationKind();
			}
			catch(Exception ex)
			{
				this.addError(ex.getMessage());
				return PointToPolygonLocationKind.NotSet;
			}
		}
	}

	private CoordinateSystem3d createLocalCoordinateSystem()
	{
		Edge3d arbitraryEdge = this.polygon.getEdge(0);
		
		Vector3d csOrigin = arbitraryEdge.vectorA();
		
		Vector3d x1 = arbitraryEdge.vectorB().substract(arbitraryEdge.vectorA(), this.precisionConfiguration).unitVector(this.precisionConfiguration);
		
		return CoordinateSystem3d.create(csOrigin, x1, this.planeNormalUnitVector, this.precisionConfiguration);
	}
	
	private Polygon3d transformInputPolygonToLocalCoordinateSystem(CoordinateSystem3d localCoordinateSystem)
	{
		Vector3d[] vertices = this.polygon.vertices();
		Vector3d[] localCoordinateSystemPolygonVertices = new Vector3d[vertices.length];
		for (int i = 0; i < vertices.length; i++)
		{
			localCoordinateSystemPolygonVertices[i] = localCoordinateSystem.toLocal(vertices[i], precisionConfiguration);
		}
		return new Polygon3d(localCoordinateSystemPolygonVertices);
	}
	
	private boolean isPolygonPlane(Polygon3d polygon)
	{
		Vector3d[] vertices = polygon.vertices();
		for	(int i = 0; i < vertices.length; i++)
		{
			if (this.precisionConfiguration.compareWithAccuracy(vertices[i].z(), 0) != 0)
			{
				return false;
			}
		}
		return true;
	}
	
	private Edge2d edge3dTo2d(Edge3d edge3d)
	{
		return new Edge2d(this.vector3dTo2d(edge3d.vectorA()), this.vector3dTo2d(edge3d.vectorB()));
	}
	
	private Vector2d vector3dTo2d(Vector3d vector3d)
	{
		return new Vector2d(vector3d.x(), vector3d.y(), this.precisionConfiguration);
	}
	
	private final boolean edgesIntersectInside(double r0, double t0)
	{
		if (this.precisionConfiguration.compareWithAccuracy(r0, 0) == 1 && this.precisionConfiguration.compareWithAccuracy(r0, 1) == -1 &&
				this.precisionConfiguration.compareWithAccuracy(t0, 0) == 1 && this.precisionConfiguration.compareWithAccuracy(t0, 1) == -1)
			return true;
		 else
			return false;
	}
}
