package spatial.algorithms.pointToPolygonLocation;

import precision.model.IPrecisionConfiguration;
import spatial.algorithms.AlgorithmBase;
import spatial.algorithms.pointToPolygonDistance.PointToPolygonDistanceAlgorithm;
import spatial.geometry.Edge3d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.pointToEdgeDistance.PointToEdgeLocationKind;
import spatial.model.algorithms.pointToPolygonDistance.IPointToPolygonDistanceAlgorithm;
import spatial.model.algorithms.pointToPolygonDistance.IPointToPolygonDistanceResult;
import spatial.model.algorithms.pointToPolygonLocation.IPointToPolygonLocationAlgorithm;
import spatial.model.algorithms.pointToPolygonLocation.IPointToPolygonLocationResult;
import spatial.model.algorithms.pointToPolygonLocation.PointToPolygonLocationKind;

public final class PointToPolygonLocationAlgorithm extends AlgorithmBase<IPointToPolygonLocationResult> implements IPointToPolygonLocationAlgorithm
{
	//start: algorithm input arguments

	private Vector3d planeNormalUnitVector;
	
	private Polygon3d polygon;
	
	private Vector3d testPoint;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private PointToPolygonLocationKind pointToPolygonLocationKind;
	
	//end: algorithm result arguments
	
	public PointToPolygonLocationAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public IPointToPolygonLocationResult calculate(Vector3d planeNormalUnitVector, Polygon3d polygon, Vector3d testPoint)
	{
		if (planeNormalUnitVector == null)
			throw new IllegalArgumentException("Plane Normal Unit Vector is not defined.");
		if (polygon == null)
			throw new IllegalArgumentException("Polygon is not defined.");
		if (testPoint == null)
			throw new IllegalArgumentException("Test Point is not defined.");
		
		this.planeNormalUnitVector = planeNormalUnitVector;
		this.polygon = polygon;
		this.testPoint = testPoint;
		
		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
//		//move input polygon to polygon local coordinate system
//		CoordinateSystem3d localCoordinateSystem = this.createLocalCoordinateSystem();
//		
//		Polygon3d polygonLocal = this.transformInputPolygonToLocalCoordinateSystem(localCoordinateSystem);
//		Vector3d testPointLocal = localCoordinateSystem.toLocal(this.testPoint, this.precisionConfiguration);
	
		IPointToPolygonDistanceAlgorithm pointToPolygonDistanceAlgorithm = new PointToPolygonDistanceAlgorithm(this.precisionConfiguration);
		IPointToPolygonDistanceResult result = pointToPolygonDistanceAlgorithm.calculate(this.planeNormalUnitVector, this.polygon, this.testPoint);
		
		if (result.hasErrors())
		{
			this.addErrors(result.errors());
			return;
		}
		
		PointToEdgeLocationKind pointLocationKind = result.distancePointLocationKind();
		int edgeIndex = result.edgeIndex();
		
		if (pointLocationKind == PointToEdgeLocationKind.InnerPoint)
		{
			Edge3d edge = this.polygon.getEdge(edgeIndex);
			
			Vector3d t = this.testPoint.substract(edge.vectorA(), this.precisionConfiguration);
			
			Vector3d v = edge.edgeVector(this.precisionConfiguration).crossProduct(t, this.precisionConfiguration);
			
			double test = this.planeNormalUnitVector.dotProduct(v);
			
			if (this.precisionConfiguration.compareWithAccuracy(test, 0) > 0)
			{
				this.pointToPolygonLocationKind = PointToPolygonLocationKind.Inside;
			}
			else if (this.precisionConfiguration.compareWithAccuracy(test, 0) == 0)
			{
				this.pointToPolygonLocationKind = PointToPolygonLocationKind.OnBoundary;
			}
			else
			{
				this.pointToPolygonLocationKind = PointToPolygonLocationKind.Outside;
			}
		}
		else if (pointLocationKind == PointToEdgeLocationKind.EndPoint || pointLocationKind == PointToEdgeLocationKind.StartPoint)
		{
			Edge3d edgeA = this.polygon.getEdge(edgeIndex);
			Edge3d edgeB = this.polygon.getEdge((edgeIndex + 1) % this.polygon.vertices().length);

			Vector3d a = edgeA.edgeVector(this.precisionConfiguration);
			Vector3d b = edgeB.edgeVector(this.precisionConfiguration);
			
			//h
			Vector3d axb = a.crossProduct(b, this.precisionConfiguration);
			double h = this.planeNormalUnitVector.dotProduct(axb);
			
			//ak
			Vector3d ta = this.testPoint.substract(edgeA.vectorA(), this.precisionConfiguration);
			double aa = this.planeNormalUnitVector.dotProduct(a.crossProduct(ta, this.precisionConfiguration));

			if (this.precisionConfiguration.compareWithAccuracy(h, 0) != 0)
			{
				//ak+1
				Vector3d tb = this.testPoint.substract(edgeB.vectorA(), this.precisionConfiguration);
				double ab = this.planeNormalUnitVector.dotProduct(b.crossProduct(tb, this.precisionConfiguration));
				
				if (this.precisionConfiguration.compareWithAccuracy(h, 0) > 0)
				{
					if (this.precisionConfiguration.compareWithAccuracy(aa, 0) > 0 && this.precisionConfiguration.compareWithAccuracy(ab, 0) > 0)
					{
						this.pointToPolygonLocationKind = PointToPolygonLocationKind.Inside;
					}
					else
					{
						this.pointToPolygonLocationKind = PointToPolygonLocationKind.Outside;
					}
				}
				else
				{
					if (this.precisionConfiguration.compareWithAccuracy(aa, 0) < 0 && this.precisionConfiguration.compareWithAccuracy(ab, 0) < 0)
					{
						this.pointToPolygonLocationKind = PointToPolygonLocationKind.Outside;
					} 
					else
					{
						this.pointToPolygonLocationKind = PointToPolygonLocationKind.Inside;
					}
				}
			}
			else
			{
				if (this.precisionConfiguration.compareWithAccuracy(aa, 0) > 0)
				{
					this.pointToPolygonLocationKind = PointToPolygonLocationKind.Inside;
				}
				else
				{
					this.pointToPolygonLocationKind = PointToPolygonLocationKind.Outside;
				}
			}
		}
		else
		{
			this.addError("Unexpected point to polygon location kind.");
		}		
	}

	@Override
	protected IPointToPolygonLocationResult constructResult()
	{
		return new PointToPolygonLocationResult(this.pointToPolygonLocationKind);
	}

	@Override
	protected IPointToPolygonLocationResult constructErrorsResult()
	{
		return new PointToPolygonLocationResult(this.errors());
	}

}
