package spatial.algorithms.polygon2dSelfIntersection;

import java.util.Arrays;
import java.util.Comparator;

import precision.model.IPrecisionConfiguration;
import precision.model.PreciseDouble;
import spatial.algorithms.AlgorithmBase;
import spatial.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationAlgorithm;
import spatial.geometry.CoordinateSystem3d;
import spatial.geometry.Edge2d;
import spatial.geometry.Edge3d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector2d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationAlgorithm;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionAlgorithm;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionResult;
import spatial.model.algorithms.polygon2dSelfIntersection.Polygon2dSelfIntersectionResultCode;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionResult.IntersectedPair;

public final class Polygon2dSelfIntersectionAlgorithm extends AlgorithmBase<IPolygon2dSelfIntersectionResult> implements IPolygon2dSelfIntersectionAlgorithm
{
	//start: algorithm input arguments
	
	private Vector3d planeNormalUnitVector;
	
	private Polygon3d polygon;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private Polygon2dSelfIntersectionResultCode resultCode;
	
	private IntersectedPair intersectedPair;
	
	//end: algorithm result arguments
	
	public Polygon2dSelfIntersectionAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public final IPolygon2dSelfIntersectionResult calculate(Vector3d planeNormalUnitVector, Polygon3d polygon)
	{
		if (planeNormalUnitVector == null)
			throw new IllegalArgumentException("The Plane Normal Unit Vector is not defined.");
		if (polygon == null)
			throw new IllegalArgumentException("The Polygon is not defined.");
		
		this.planeNormalUnitVector = planeNormalUnitVector;
		this.polygon = polygon;
		
		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
		//move input polygon to polygon local coordinate system
		CoordinateSystem3d localCoordinateSystem = this.createLocalCoordinateSystem();
		Polygon3d localCsPolygon = this.transformInputPolygonToLocalCoordinateSystem(localCoordinateSystem);
		
		//check local polygon is plane
		Vector3d[] localVertices = localCsPolygon.vertices();
		for	(int i = 0; i < localVertices.length; i++)
		{
			if (this.precisionConfiguration.compareWithAccuracy(localVertices[i].z(), 0) != 0)
			{
				this.addError("Polygon isn't plane.");
				return;
			}
		}

		// create sorted array of EdgeCoor
		EdgeCoor[] ecoors = new EdgeCoor[localVertices.length];
		for (int i = 0; i < ecoors.length; i++)
			ecoors[i] = new EdgeCoor(localCsPolygon.getEdge(i), i);
		Arrays.sort(ecoors, new EdgeCoorComparator());
		
		// sweep line algorithm
		for (int i = 0; i < ecoors.length; i++)
		{
			EdgeCoor ei = ecoors[i];
			int touchesInBoundaryNode = 0;
			for (int j = i + 1; j < ecoors.length; j++)
			{
				EdgeCoor ej = ecoors[j];
				
				if (ej.minX1.compareTo(ei.maxX1) > 0 && ej.minX2.compareTo(ei.maxX2) > 0 && ej.maxX2.compareTo(ei.minX2) < 0)
					continue;
				
				EdgeIntersectionKind xKind = this.testIntersection(ei.edge, ej.edge);
				
				if (xKind == null)
				{
					this.addError("Couldn't run intersection test.");
					return;
				}
				
				if (xKind == EdgeIntersectionKind.AtNode)
				{
					touchesInBoundaryNode++;
					if (touchesInBoundaryNode > 2)
					{
						this.resultCode = Polygon2dSelfIntersectionResultCode.SelfIntersected;
						
						this.intersectedPair = new IntersectedPair(ei.index, ej.index);
						return;
					}
				} 
				else if (xKind == EdgeIntersectionKind.Inside)
				{
					this.resultCode = Polygon2dSelfIntersectionResultCode.SelfIntersected;
					
					this.intersectedPair = new IntersectedPair(ei.index, ej.index);
					return;
				}
			}
		}
		
		this.resultCode = Polygon2dSelfIntersectionResultCode.NotSelfIntersected;		
	}

	@Override
	protected IPolygon2dSelfIntersectionResult constructResult()
	{
		return new Polygon2dSelfIntersectionResult(this.resultCode, this.intersectedPair);
	}

	@Override
	protected IPolygon2dSelfIntersectionResult constructErrorsResult()
	{
		return new Polygon2dSelfIntersectionResult(this.errors());
	}

	private Polygon3d transformInputPolygonToLocalCoordinateSystem(CoordinateSystem3d localCoordinateSystem)
	{
		Vector3d[] vertices = this.polygon.vertices();
		Vector3d[] localCoordinateSystemPolygonVertices = new Vector3d[vertices.length];
		for (int i = 0; i < vertices.length; i++)
		{
			localCoordinateSystemPolygonVertices[i] = localCoordinateSystem.toLocal(vertices[i], precisionConfiguration);
		}
		
		return new Polygon3d(localCoordinateSystemPolygonVertices);
	}

	private CoordinateSystem3d createLocalCoordinateSystem()
	{
		Edge3d arbitraryEdge = this.polygon.getEdge(0);
		
		Vector3d csOrigin = arbitraryEdge.vectorA();
		
		Vector3d x1 = arbitraryEdge.vectorB().substract(arbitraryEdge.vectorA(), this.precisionConfiguration).unitVector(this.precisionConfiguration);
		
		return CoordinateSystem3d.create(csOrigin, x1, this.planeNormalUnitVector, this.precisionConfiguration);
	}
	
	private EdgeIntersectionKind testIntersection(Edge3d e1, Edge3d e2)
	{
		IEdgeToEdgeLocationAlgorithm alg = new EdgeToEdgeLocationAlgorithm(this.precisionConfiguration);
		IEdgeToEdgeLocationResult result = alg.calculate(this.edge3dTo2d(e1), this.edge3dTo2d(e2));
		if (result.hasErrors())
		{
			this.addErrors(result.errors());
			return null;
		}

		if (result.edgesIntersectAtNode())
			return EdgeIntersectionKind.AtNode;
		else if (result.edgesIntersectInside())
			return EdgeIntersectionKind.Inside;
		else
			return EdgeIntersectionKind.NotIntersect;
	}
	
	private Edge2d edge3dTo2d(Edge3d edge3d)
	{
		return new Edge2d(new Vector2d(edge3d.vectorA().preciseX(), edge3d.vectorA().preciseY()),
				new Vector2d(edge3d.vectorB().preciseX(), edge3d.vectorB().preciseY()));
	}
	
	class EdgeCoor
	{
		public final Edge3d edge;
		
		public final int index;
		
		public final PreciseDouble minX1;
		
		public final PreciseDouble maxX1;
		
		public final PreciseDouble minX2;
		
		public final PreciseDouble maxX2;
		
		public EdgeCoor(Edge3d edge, int index)
		{
			this.edge = edge;
			this.index = index;
			
			int xcmp = edge.vectorA().preciseX().compareTo(edge.vectorB().preciseX());
			if (xcmp == 1)
			{
				this.minX1 = edge.vectorB().preciseX();
				this.maxX1 = edge.vectorA().preciseX();
			} 
			else
			{
				this.minX1 = edge.vectorA().preciseX();
				this.maxX1 = edge.vectorB().preciseX();
			}
			
			int ycmp = edge.vectorA().preciseY().compareTo(edge.vectorB().preciseY());
			if (ycmp == 1)
			{
				this.minX2 = edge.vectorB().preciseY();
				this.maxX2 = edge.vectorA().preciseY();
			} 
			else
			{
				this.minX2 = edge.vectorA().preciseY();
				this.maxX2 = edge.vectorB().preciseY();
			}
		}
		
	}

	class EdgeCoorComparator implements Comparator<EdgeCoor> 
	{
	    @Override
	    public int compare(EdgeCoor ec1, EdgeCoor ec2) 
	    {
	        return ec1.minX1.compareTo(ec2.minX1);
	    }
	}

	enum EdgeIntersectionKind
	{
		AtNode,
		Inside,
		NotIntersect
	}
}