package spatial.algorithms.polygonToPolygonLocation;

import java.util.Arrays;
import java.util.Comparator;

import precision.model.IPrecisionConfiguration;
import precision.model.PreciseDouble;
import spatial.algorithms.AlgorithmBase;
import spatial.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationAlgorithm;
import spatial.geometry.CoordinateSystem3d;
import spatial.geometry.Edge2d;
import spatial.geometry.Edge3d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector2d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationAlgorithm;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationAlgorithm;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationResult;
import spatial.model.algorithms.polygonToPolygonLocation.PolygonToPolygonLocationKind;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationResult.IntersectedEdgesPair;

public final class PolygonToPolygonLocationAlgorithm extends AlgorithmBase<IPolygonToPolygonLocationResult> implements IPolygonToPolygonLocationAlgorithm
{
	//start: algorithm input arguments
	
	private Vector3d planeNormalUnitVector;
	
	private Polygon3d aPolygon;
	
	private Polygon3d bPolygon;

	//end: algorithm input arguments
	
	//start: algorithm result arguments
	
	private PolygonToPolygonLocationKind locationKind;
	
	private IntersectedEdgesPair intersectedEdgesPair;
	
	//end: algorithm result arguments
	
	public PolygonToPolygonLocationAlgorithm(IPrecisionConfiguration precisionConfiguration)
	{
		super(precisionConfiguration);
	}

	@Override
	public final IPolygonToPolygonLocationResult calculate(Vector3d planeNormalUnitVector, Polygon3d polygon1, Polygon3d polygon2)
	{
		if (planeNormalUnitVector == null)
			throw new IllegalArgumentException("The Plane Normal Unit Vector is not defined.");
		if (polygon1 == null)
			throw new IllegalArgumentException("The Polygon 1 is not defined.");
		if (polygon2 == null)
			throw new IllegalArgumentException("The Polygon 2 is not defined.");
		
		this.planeNormalUnitVector = planeNormalUnitVector;
		this.aPolygon = polygon1;
		this.bPolygon = polygon2;
		
		return super.calculate();
	}

	@Override
	protected void calculateInternal()
	{
		//move input polygon to polygon local coordinate system
		CoordinateSystem3d localCoordinateSystem = this.createLocalCoordinateSystem();
		Polygon3d aPolygonLocal = this.transformInputPolygonToLocalCoordinateSystem(localCoordinateSystem, this.aPolygon);
		Polygon3d bPolygonLocal = this.transformInputPolygonToLocalCoordinateSystem(localCoordinateSystem, this.bPolygon);
		
		//check local polygons are plane
		if (!this.isPolygonPlane(aPolygonLocal))
		{
			this.addError("Polygon A isn't plane.");
			return;
		}
		if (!this.isPolygonPlane(bPolygonLocal))
		{
			this.addError("Polygon B isn't plane.");
			return;
		}
		
		// create sorted array of EdgeCoor
		EdgeCoor[] ecoors = new EdgeCoor[aPolygonLocal.vertices().length + bPolygonLocal.vertices().length];
		for (int i = 0; i < aPolygonLocal.vertices().length; i++)
			ecoors[i] = new EdgeCoor(aPolygonLocal.getEdge(i), i, "a");
		for (int i = 0; i < bPolygonLocal.vertices().length; i++)
			ecoors[aPolygonLocal.vertices().length + i] = new EdgeCoor(bPolygonLocal.getEdge(i), i, "b");
		Arrays.sort(ecoors, new EdgeCoorComparator());
		
		// sweep line algorithm with the colored edges
		for (int i = 0; i < ecoors.length; i++)
		{
			EdgeCoor ei = ecoors[i];

			for (int j = i + 1; j < ecoors.length; j++)
			{
				EdgeCoor ej = ecoors[j];
				
				if (ei.polygonName.equals(ej.polygonName))
					continue;
				
				if (ej.minX1.compareTo(ei.maxX1) > 0 && ej.minX2.compareTo(ei.maxX2) > 0 && ej.maxX2.compareTo(ei.minX2) < 0)
					continue;
				
				EdgeIntersectionKind xKind = this.testIntersection(ei.edge, ej.edge);
				
				if (xKind == null)
				{
					this.addError("Couldn't run intersection test.");
					return;
				}
				
				if (xKind == EdgeIntersectionKind.AtNode || xKind == EdgeIntersectionKind.Inside)
				{
					this.locationKind = PolygonToPolygonLocationKind.Intersects;
					
					if (ei.polygonName.equals("a"))
						this.intersectedEdgesPair = new IntersectedEdgesPair(ei.index, ej.index);
					else
						this.intersectedEdgesPair = new IntersectedEdgesPair(ej.index, ei.index);
					
					return;
					
				}
			}
		}
		
		this.locationKind = PolygonToPolygonLocationKind.NotIntersects;
		
	}

	@Override
	protected IPolygonToPolygonLocationResult constructResult()
	{
		return new PolygonToPolygonLocationResult(this.locationKind, this.intersectedEdgesPair);
	}

	@Override
	protected IPolygonToPolygonLocationResult constructErrorsResult()
	{
		return new PolygonToPolygonLocationResult(this.errors());
	}

	private CoordinateSystem3d createLocalCoordinateSystem()
	{
		Edge3d arbitraryEdge = this.aPolygon.getEdge(0);
		
		Vector3d csOrigin = arbitraryEdge.vectorA();
		
		Vector3d x1 = arbitraryEdge.vectorB().substract(arbitraryEdge.vectorA(), this.precisionConfiguration).unitVector(this.precisionConfiguration);
		
		return CoordinateSystem3d.create(csOrigin, x1, this.planeNormalUnitVector, this.precisionConfiguration);
	}
	
	private Polygon3d transformInputPolygonToLocalCoordinateSystem(CoordinateSystem3d localCoordinateSystem, Polygon3d polygon)
	{
		Vector3d[] vertices = polygon.vertices();
		Vector3d[] localCoordinateSystemPolygonVertices = new Vector3d[vertices.length];
		for (int i = 0; i < vertices.length; i++)
		{
			localCoordinateSystemPolygonVertices[i] = localCoordinateSystem.toLocal(vertices[i], precisionConfiguration);
		}
		
		return new Polygon3d(localCoordinateSystemPolygonVertices);
	}
	
	private boolean isPolygonPlane(Polygon3d polygon)
	{
		Vector3d[] vertices = polygon.vertices();
		for	(int i = 0; i < vertices.length; i++)
		{
			if (this.precisionConfiguration.compareWithAccuracy(vertices[i].z(), 0) != 0)
			{
				return false;
			}
		}
		return true;
	}
	
	private Edge2d edge3dTo2d(Edge3d edge3d)
	{
		return new Edge2d(new Vector2d(edge3d.vectorA().preciseX(), edge3d.vectorA().preciseY()),
				new Vector2d(edge3d.vectorB().preciseX(), edge3d.vectorB().preciseY()));
	}
	
	private EdgeIntersectionKind testIntersection(Edge3d e1, Edge3d e2)
	{
		IEdgeToEdgeLocationAlgorithm alg = new EdgeToEdgeLocationAlgorithm(this.precisionConfiguration);
		IEdgeToEdgeLocationResult result = alg.calculate(this.edge3dTo2d(e1), this.edge3dTo2d(e2));
		
		if (result.hasErrors())
		{
			this.addErrors(result.errors());
			return null;
		}

		if (result.edgesIntersectAtNode())
			return EdgeIntersectionKind.AtNode;
		else if (result.edgesIntersectInside())
			return EdgeIntersectionKind.Inside;
		else
			return EdgeIntersectionKind.NotIntersect;
	}

	class EdgeCoor
	{
		public final Edge3d edge;
		
		public final int index;
		
		public final String polygonName;
		
		public final PreciseDouble minX1;
		
		public final PreciseDouble maxX1;
		
		public final PreciseDouble minX2;
		
		public final PreciseDouble maxX2;
		
		public EdgeCoor(Edge3d edge, int index, String polygonName)
		{
			this.edge = edge;
			this.index = index;
			this.polygonName = polygonName;
			
			int xcmp = edge.vectorA().preciseX().compareTo(edge.vectorB().preciseX());
			if (xcmp == 1)
			{
				this.minX1 = edge.vectorB().preciseX();
				this.maxX1 = edge.vectorA().preciseX();
			} 
			else
			{
				this.minX1 = edge.vectorA().preciseX();
				this.maxX1 = edge.vectorB().preciseX();
			}
			
			int ycmp = edge.vectorA().preciseY().compareTo(edge.vectorB().preciseY());
			if (ycmp == 1)
			{
				this.minX2 = edge.vectorB().preciseY();
				this.maxX2 = edge.vectorA().preciseY();
			} 
			else
			{
				this.minX2 = edge.vectorA().preciseY();
				this.maxX2 = edge.vectorB().preciseY();
			}
		}
		
	}

	class EdgeCoorComparator implements Comparator<EdgeCoor> 
	{
	    @Override
	    public int compare(EdgeCoor ec1, EdgeCoor ec2) 
	    {
	        return ec1.minX1.compareTo(ec2.minX1);
	    }
	}

	enum EdgeIntersectionKind
	{
		AtNode,
		Inside,
		NotIntersect
	}
}
