package spatial.geometry;

import org.ejml.data.DMatrixRMaj;
import org.ejml.dense.row.CommonOps_DDRM;

import precision.model.IPrecisionConfiguration;
import spatial.model.IGeometry;

public class CoordinateSystem3d implements IGeometry
{
	private final Vector3d origin;
	
	private final Vector3d e1;
	
	private final Vector3d e2;
	
	private final Vector3d e3;
	
	private DMatrixRMaj coordinateSystemBasisMatrix;
	
	public CoordinateSystem3d(Vector3d origin, Vector3d e1, Vector3d e2, Vector3d e3)
	{
		if (origin == null)
			throw new IllegalArgumentException("Origin is not defined.");
		if (e1 == null)
			throw new IllegalArgumentException("E1 axis is not defined.");
		if (e2 == null)
			throw new IllegalArgumentException("E2 axis is not defined.");
		if (e3 == null)
			throw new IllegalArgumentException("E3 axis is not defined.");
		
		//TODO: vectors have to be of unit length and orthogonal
		
		this.origin = origin;
		this.e1 = e1;
		this.e2 = e2;
		this.e3 = e3;
	}
	
	public static CoordinateSystem3d create(Vector3d origin, Vector3d x1, Vector3d x3, IPrecisionConfiguration precisionConfiguration)
	{
		Vector3d x2 = x3.crossProduct(x1, precisionConfiguration).unitVector(precisionConfiguration);
		
		return new CoordinateSystem3d(origin, x1, x2, x3);
	}
	
	public final Vector3d origin()
	{
		return this.origin;
	}
	
	public final Vector3d e1()
	{
		return this.e1;
	}
	
	public final Vector3d e2()
	{
		return this.e2;
	}
	
	public final Vector3d e3()
	{
		return this.e3;
	}
	
	public final Vector3d toLocal(Vector3d globalPoint, IPrecisionConfiguration precisionConfiguration)
	{
		return this.toLocal(new Vector3d[] {globalPoint}, precisionConfiguration)[0];
	}
	
	public final Vector3d[] toLocal(Vector3d[] globalPoints, IPrecisionConfiguration precisionConfiguration)
	{
		this.ensureCoordinateSystemBasisMatrix();
		
		//compute local points
		Vector3d[] localPoints = new Vector3d[globalPoints.length];
		DMatrixRMaj lp = new DMatrixRMaj(3, 1);
		for (int i = 0; i < globalPoints.length; i++)
		{
			Vector3d shiftedGp = globalPoints[i].substract(this.origin, precisionConfiguration);
			
			CommonOps_DDRM.mult(this.coordinateSystemBasisMatrix, DMatrixRMaj.wrap(3, 1, new double[] {shiftedGp.x(), shiftedGp.y(), shiftedGp.z()}), lp);
			localPoints[i] = new Vector3d(lp.get(0, 0), lp.get(1, 0), lp.get(2, 0), precisionConfiguration);
		}
		
		return localPoints;
	}
	
	public final Vector3d toGlobal(Vector3d localPoint, IPrecisionConfiguration precisionConfiguration)
	{
		return this.toGlobal(new Vector3d[] {localPoint}, precisionConfiguration)[0];
	}
	
	public final Vector3d[] toGlobal(Vector3d[] localPoints, IPrecisionConfiguration precisionConfiguration)
	{
		this.ensureCoordinateSystemBasisMatrix();
		
		DMatrixRMaj transposedCoordinateSystemBasisMatrix = new DMatrixRMaj(3, 3);
		CommonOps_DDRM.transpose(this.coordinateSystemBasisMatrix, transposedCoordinateSystemBasisMatrix);
		
		//compute local points
		Vector3d[] globalPoints = new Vector3d[localPoints.length];
		DMatrixRMaj shiftedGlobalPoint = new DMatrixRMaj(3, 1);
		for (int i = 0; i < localPoints.length; i++)
		{
			CommonOps_DDRM.mult(transposedCoordinateSystemBasisMatrix, DMatrixRMaj.wrap(3, 1, new double[] {localPoints[i].x(), localPoints[i].y(), localPoints[i].z()}), shiftedGlobalPoint);
			
			Vector3d globalPoint = new Vector3d(shiftedGlobalPoint.get(0, 0) + this.origin.x(), 
					shiftedGlobalPoint.get(1, 0)+ this.origin.y(), 
					shiftedGlobalPoint.get(2, 0)+ this.origin.z(), 
					precisionConfiguration);
			
			globalPoints[i] = globalPoint;
		}
		
		return globalPoints;
	}
	
	public final DMatrixRMaj coordinateSystemBasisMatrix()
	{
		this.ensureCoordinateSystemBasisMatrix();
		
		return this.coordinateSystemBasisMatrix;
	}

	private final void ensureCoordinateSystemBasisMatrix()
	{
		if (this.coordinateSystemBasisMatrix == null)
		{
			DMatrixRMaj mx = new DMatrixRMaj(3, 3);

			mx.set(0, 0, this.e1.x());
			mx.set(1, 0, this.e1.y());
			mx.set(2, 0, this.e1.z());
			
			mx.set(0, 1, this.e2.x());
			mx.set(1, 1, this.e2.y());
			mx.set(2, 1, this.e2.z());
			
			mx.set(0, 2, this.e3.x());
			mx.set(1, 2, this.e3.y());
			mx.set(2, 2, this.e3.z());
			
			this.coordinateSystemBasisMatrix = mx;
		}
	}
}
