package spatial.geometry;

import precision.model.IPreciseDoubleFactory;
import precision.model.IPrecisionConfiguration;
import precision.model.PreciseDouble;
import spatial.model.IGeometry;

public class Vector2d implements IGeometry
{

	private final PreciseDouble ax;
	
	private final PreciseDouble ay;
	
	private int accurateHashCode;
	
	public Vector2d(double x, double y, IPreciseDoubleFactory preciseDoubleFactory)
	{	
		this.ax = preciseDoubleFactory.create(x);
		this.ay = preciseDoubleFactory.create(y);
	}
	
	public Vector2d(double x, double y, IPrecisionConfiguration precisionConfiguration)
	{
		this.ax = new PreciseDouble(x, precisionConfiguration);
		this.ay = new PreciseDouble(y, precisionConfiguration);
	}
	
	public Vector2d(PreciseDouble x, PreciseDouble y)
	{	
		this.ax = x;
		this.ay = y;
	}
	
	public final double x()
	{
		return this.ax.rawDouble();
	}
	
	public final double y()
	{
		return this.ay.rawDouble();
	}
	
	public final PreciseDouble preciseX()
	{
		return this.ax;
	}
	
	public final PreciseDouble preciseY()
	{
		return this.ay;
	}
	
	public final Vector2d substract(Vector2d v, IPrecisionConfiguration precisionConfiguration)
	{
		return new Vector2d(this.ax.rawDouble() - v.ax.rawDouble(), 
				this.ay.rawDouble() - v.ay.rawDouble(), precisionConfiguration);
	}
	
	public final Vector2d add(Vector2d v, IPrecisionConfiguration precisionConfiguration)
	{
		return new Vector2d(this.x() + v.x(), this.y() + v.y(), precisionConfiguration);
	}
	
	public final double dotProduct(Vector2d v)
	{
		return this.x() * v.x() + this.y() * v.y();
	}
	
	public final Vector2d mul(double k, IPrecisionConfiguration precisionConfiguration)
	{
		return new Vector2d(this.x() * k, this.y() * k, precisionConfiguration);
	}

	public final Vector2d unitVector(IPrecisionConfiguration precisionConfiguration)
	{
		double eNorm = this.euqlideanNorm();
		return new Vector2d(this.x() / eNorm, this.y() / eNorm, precisionConfiguration);
	}
	
	@Override
	public int hashCode()
	{
		if (this.accurateHashCode == 0)
		{
			this.accurateHashCode = (41 * 43 + this.ax.hashCode()) * 43 + this.ay.hashCode();
		}
		return this.accurateHashCode;
	}
	
	@Override
	public boolean equals(Object other)
	{
		if (other == null || this.getClass() != other.getClass())
			return false;
		if (this == other)
			return true;
		Vector2d otherVector = (Vector2d) other;

		return this.ax.equals(otherVector.ax) && this.ay.equals(otherVector.ay);
	}
	
	@Override
	public String toString()
	{
		return String.format("%f; %f", this.ax.rawDouble(), this.ay.rawDouble());
	}

	public boolean isUnit(IPrecisionConfiguration precisionConfiguration) 
	{
		return Math.abs(1 - this.euqlideanNorm()) < precisionConfiguration.distanceDelta();
	}
	
	public double euqlideanNorm()
	{
		return Math.sqrt(this.x() * this.x() + this.y() * this.y());
	}

}
