package spatial.geometry;

import precision.model.IPreciseDoubleFactory;
import precision.model.IPrecisionConfiguration;
import precision.model.PreciseDouble;
import spatial.model.IGeometry;

public final class Vector3d implements IGeometry 
{
	
	private final PreciseDouble ax;
	
	private final PreciseDouble ay;
	
	private final PreciseDouble az;
	
	private int accurateHashCode;
	
	public Vector3d(double x, double y, double z, IPreciseDoubleFactory preciseDoubleFactory)
	{	
		this.ax = preciseDoubleFactory.create(x);
		this.ay = preciseDoubleFactory.create(y);
		this.az = preciseDoubleFactory.create(z);
	}
	
	public Vector3d(double x, double y, double z, IPrecisionConfiguration precisionConfiguration)
	{
		this.ax = new PreciseDouble(x, precisionConfiguration);
		this.ay = new PreciseDouble(y, precisionConfiguration);
		this.az = new PreciseDouble(z, precisionConfiguration);
	}
	
	public Vector3d(PreciseDouble x, PreciseDouble y, PreciseDouble z)
	{	
		this.ax = x;
		this.ay = y;
		this.az = z;
	}
	
	public final double x()
	{
		return this.ax.rawDouble();
	}
	
	public final double y()
	{
		return this.ay.rawDouble();
	}
	
	public final double z()
	{
		return this.az.rawDouble();
	}
	
	public final PreciseDouble preciseX()
	{
		return this.ax;
	}
	
	public final PreciseDouble preciseY()
	{
		return this.ay;
	}
	
	public final PreciseDouble preciseZ()
	{
		return this.az;
	}
	
	public final Vector3d substract(Vector3d v, IPrecisionConfiguration precisionConfiguration)
	{
		return new Vector3d(this.ax.rawDouble() - v.ax.rawDouble(), 
				this.ay.rawDouble() - v.ay.rawDouble(), this.az.rawDouble() - v.az.rawDouble(), precisionConfiguration);
	}
	
	public final Vector3d add(Vector3d v, IPrecisionConfiguration precisionConfiguration)
	{
		return new Vector3d(this.x() + v.x(), this.y() + v.y(), this.z() + v.z(), precisionConfiguration);
	}
	
	public final double dotProduct(Vector3d v)
	{
		return this.x() * v.x() + this.y() * v.y() + this.z() * v.z();
	}
	
	public final Vector3d mul(double k, IPrecisionConfiguration precisionConfiguration)
	{
		return new Vector3d(this.x() * k, this.y() * k, this.z() * k, precisionConfiguration);
	}
	
	public final Vector3d crossProduct(Vector3d v, IPrecisionConfiguration precisionConfiguration) 
	{
        return new Vector3d(this.y() * v.z() - this.z() * v.y(),
        		-(this.x() * v.z() - this.z() * v.x()),
        		this.x() * v.y() - this.y() * v.x(),
        		precisionConfiguration);
    }
	
	public final Vector3d unitVector(IPrecisionConfiguration precisionConfiguration)
	{
		double eNorm = this.euqlideanNorm();
		return new Vector3d(this.x() / eNorm, this.y() / eNorm, this.z() / eNorm, precisionConfiguration);
	}
	
	@Override
	public int hashCode()
	{
		if (this.accurateHashCode == 0)
		{
			this.accurateHashCode = ((41 * 43 + this.ax.hashCode()) * 43 + this.ay.hashCode()) * 43 + this.az.hashCode();
		}
		return this.accurateHashCode;
	}
	
	@Override
	public boolean equals(Object other)
	{
		if (other == null || this.getClass() != other.getClass())
			return false;
		if (this == other)
			return true;
		Vector3d otherVector = (Vector3d) other;

		return this.ax.equals(otherVector.ax) && this.ay.equals(otherVector.ay) && this.az.equals(otherVector.az);
	}
	
	@Override
	public String toString()
	{
		return String.format("%f; %f; %f", this.ax.rawDouble(), this.ay.rawDouble(), this.az.rawDouble());
	}

	public boolean isUnit(IPrecisionConfiguration precisionConfiguration) 
	{
		return Math.abs(1 - this.euqlideanNorm()) < precisionConfiguration.distanceDelta();
	}
	
	public double euqlideanNorm()
	{
		return Math.sqrt(this.x() * this.x() + this.y() * this.y() + this.z() * this.z());
	}

}
