package spatial.model.algorithms;

import spatial.geometry.Edge2d;
import spatial.geometry.Edge3d;
import spatial.geometry.Face2d;
import spatial.geometry.Face3d;
import spatial.geometry.Line3d;
import spatial.geometry.Plane3d;
import spatial.geometry.Polygon2d;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.bestFittingPlane.IBestFittingPlaneResult;
import spatial.model.algorithms.csTransformation.ICsTransformationResult;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;
import spatial.model.algorithms.edgeToFaceLocation.IEdgeToFaceLocationResult;
import spatial.model.algorithms.edgeToPolygonLocation.IEdgeToPolygonLocationResult;
import spatial.model.algorithms.faceArea.IFaceAreaResult;
import spatial.model.algorithms.pointToEdgeDistance.IPointToEdgeDistanceResult;
import spatial.model.algorithms.pointToFaceLocation.IPointToFaceLocationResult;
import spatial.model.algorithms.pointToLineLocation.IPointToLineLocationResult;
import spatial.model.algorithms.pointToPlaneDistance.IPointToPlaneDistanceResult;
import spatial.model.algorithms.pointToPolygonDistance.IPointToPolygonDistanceResult;
import spatial.model.algorithms.pointToPolygonLocation.IPointToPolygonLocationResult;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionResult;
import spatial.model.algorithms.polygonArea.IPolygonAreaResult;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationResult;

/**
 * Base geometric algorithms.
 */
public interface ICommonGeometryAlgorithms
{
	/**
	 * Determine the plane which minimises the sum of the squares of the normal distances of a given point set from this plane
	 * @param points The unit normal vector n of the plane.
	 */
	IBestFittingPlaneResult bestFittingPlane(Vector3d[] points);
	
	/**
	 * Determine a local coordinate system in a plane and compute the local coordinates of a set of points, whose global coordinates are given.
	 * @param normalPlaneOrt The unit normal vector n of the plane (x3).
	 * @param inPlaneOrt The unit vector u of the plane (in-plane x1).
	 * @param localCSOrigin The Local Coordinate System reference point in the plane.
	 * @param globalCsPoints The set of points with coordinates in the global CS. 
	 */
	ICsTransformationResult csTransformation(Vector3d normalPlaneOrt, Vector3d inPlaneOrt, Vector3d localCSOrigin, Vector3d[] globalCsPoints);
	
	/**
	 * Determine the local coordinates of the point of intersection of two lines 
	 * containing given edges AB and CD, and the location of the point of intersection 
	 * relative to edges AB and CD
	 * @param edge1 The edge 1 in 2D.
	 * @param edge2 The edge 2 in 2D.
	 */
	IEdgeToEdgeLocationResult edgeToEdgeLocation(Edge2d edge1, Edge2d edge2);
	
	/**
	 * A plane with unit normal vector n contains an oriented face and a test edge E with nodes A and B. 
	 * The face is defined by an outer polygon and an optional number of inner polygons. 
	 * The location of edge E relative to the face is to be determined.
	 * @param planeNormalUnitVector Plane normal unit vector.
	 * @param face The face.
	 * @param testEdge The test edge.
	 */
	IEdgeToFaceLocationResult edgeToFaceLocation(Vector3d planeNormalUnitVector, Face3d face, Edge3d testEdge);
	
	/**
	 * Location of an Edge Relative to a Polygon.
	 * @param planeNormalUnitVector Plane normal unit vector.
	 * @param polygon The polygon.
	 * @param testEdge The Test Edge.
	 */
	IEdgeToPolygonLocationResult edgeToPolygonLocation(Vector3d planeNormalUnitVector, Polygon3d polygon, Edge3d testEdge);
	
	/**
	 * Calculate Area of the face.
	 * @param face The face.
	 */
	IFaceAreaResult faceArea(Face2d face);
	
	/**
	 * Determine the distance between a test node T and an edge AB as follows:
	 * - normal distance to an interior point of edge AB, or
	 * - smaller of the distances TA and TB
	 * @param edge The edge in 3D.
	 * @param point The Test point in 3D.
	 */
	IPointToEdgeDistanceResult pointToEdgeDistance(Edge3d edge, Vector3d point);
	
	/**
	 * Determine whether a test point T is an interior or an exterior point of a face F whose plane contains T.
	 * @param planeNormalUnitVector Plane normal unit vector.
	 * @param face The face.
	 * @param testPoint 3D global coordinates of the test point T 
	 */
	IPointToFaceLocationResult pointToFaceLocation(Vector3d planeNormalUnitVector, Face3d face, Vector3d testPoint);
	
	/**
	 * Determine the location of a test point T relative to a directed line.
	 * Both, the Line and the Test point, belongs to the plane.
	 * @param plainNormalVector Unit normal vector n of the plane.
	 * @param line The Line in 3D.
	 * @param testPoint The Test Point in 3d.
	 */
	IPointToLineLocationResult pointToLineLocation(Vector3d plainNormalVector, Line3d line, Vector3d testPoint);
	
	/**
	 * Determine the smallest distance from a test point T to a given plane in 3D space
	 * @param plane The plane.
	 * @param testPoint The Test Point in 3d.
	 */
	IPointToPlaneDistanceResult pointToPlaneDistance(Plane3d plane, Vector3d testPoint);
	
	/**
	 * Determine the inner point of an edge or the node of a plane polygon which is the nearest point of the boundary to a test point T in the plane.
	 * @param planeNormalUnitVector 3D global coordinates of the unit normal vector n of the plane.
	 * @param polygon Plane polygon.
	 * @param testPoint 3D global coordinates of a test point T in the plane
	 */
	IPointToPolygonDistanceResult pointToPolygonDistanceResult(Vector3d planeNormalUnitVector, Polygon3d polygon, Vector3d testPoint);
	
	/**
	 * A plane with unit normal vector n contains an oriented polygon and a test point T. 
	 * It is to be determined whether T is an interior or an exterior point of the polygon.
	 * @param planeNormalUnitVector 3D global coordinates of the unit normal vector n of the plane.
	 * @param polygon 3D Polygon P.
	 * @param testPoint 3D global coordinates of a test point T in the plane
	 * @return
	 */
	IPointToPolygonLocationResult pointToPolygonLocation(Vector3d planeNormalUnitVector, Polygon3d polygon, Vector3d testPoint);
	
	/**
	 * Determine whether any edges of a polygon intersect.
	 * @param planeNormal 3D global coordinates of the unit normal vector n of the plane.
	 * @param polygon Plane polygon with nodes coordinates in 3D.
	 */
	IPolygon2dSelfIntersectionResult polygon2dSelfIntersection(Vector3d planeNormal, Polygon3d polygon);
	
	/**
	 * Calculate Area of the polygon.
	 * @param polygon The polygon.
	 */
	IPolygonAreaResult polygonArea(Polygon2d polygon);
	
	/**
	 * Determine whether at least one edge of polygon P1 intersects at least one edge of polygon P2.
	 * @param planeNormalUnitVector 3D global coordinates of the unit normal vector n of the plane.
	 * @param polygon1 The polygon 1.
	 * @param polygon2 The polygon 2.
	 */
	IPolygonToPolygonLocationResult polygonToPolygonLocation(Vector3d planeNormalUnitVector, Polygon3d polygon1, Polygon3d polygon2);
	
}

