package spatial.model.elements.set;

import java.util.HashMap;

import spatial.model.IElement;
import spatial.model.IGeometry;
import spatial.model.elements.set.exception.ElementGeometryOfNotValidTypeException;
import spatial.model.elements.set.exception.ElementWithTheSameGeometryAlreadyExistsException;
import spatial.model.elements.set.exception.ElementWithTheSameNameAlreadyExistsException;

public abstract class ElementsSetBase implements IElementSet
{
	/**
	 * Collection of the elements
	 */
	private HashMap<Integer, CollidedElementsContainer> geometryToElement;
	
	private HashMap<String, IElement> nameToElement;
	
	public ElementsSetBase()
	{
		this.geometryToElement = new HashMap<Integer, CollidedElementsContainer>();
		this.nameToElement = new HashMap<String, IElement>();
	}

	@SuppressWarnings("rawtypes")
	protected abstract Class geometryClass();

	@Override
	public IElement ensure(IElement element)
			throws ElementWithTheSameNameAlreadyExistsException, 
			ElementWithTheSameGeometryAlreadyExistsException, 
			ElementGeometryOfNotValidTypeException 
	{
		if (this.geometryClass() != element.geometry().getClass())
			throw new ElementGeometryOfNotValidTypeException(this.geometryClass().getName(), element.geometry().getClass().getName());
		
		//check if the node already exists in the model
		IElement existedElement = this.nameToElement.get(element.name());
		if (existedElement != null)
		{
			if (!existedElement.geometry().equals(element.geometry()))
				throw new ElementWithTheSameNameAlreadyExistsException(element.name(), existedElement.geometry().toString(), element.geometry().toString());
			else
				return existedElement;
		}
		else
		{
			//getting the container for element by geometry hash code
			int geometryHash = element.geometry().hashCode();
			CollidedElementsContainer collidedElementsContainer = this.geometryToElement.get(geometryHash);
			
			if (collidedElementsContainer == null)
			{
				collidedElementsContainer = new CollidedElementsContainer(geometryHash, element);
				this.geometryToElement.put(geometryHash, collidedElementsContainer);
				this.nameToElement.put(element.name(), element);
			}
			else
			{
				collidedElementsContainer.addElement(element);
				this.nameToElement.put(element.name(), element);
			}
			return element;
		}
	}

	@Override
	public boolean contains(String name) 
	{
		return this.nameToElement.containsKey(name);
	}

	@Override
	public IElement get(String name) 
	{
		return this.nameToElement.get(name);
	}

	@Override
	public IElement get(IGeometry geometry) throws ElementGeometryOfNotValidTypeException 
	{
		if (this.geometryClass() != geometry.getClass())
			throw new ElementGeometryOfNotValidTypeException(this.geometryClass().getName(), geometry.getClass().getName());
		
		CollidedElementsContainer collidedElementsContainer = this.geometryToElement.get(geometry.hashCode());
		if (collidedElementsContainer != null)
		{
			return collidedElementsContainer.get(geometry);
		} 
		else
		{
			return null;
		}
	}

	@Override
	public void remove(String name) 
	{
		IElement existedNode = this.nameToElement.get(name);
		if (existedNode != null)
		{
			CollidedElementsContainer setBasket = this.geometryToElement.get(existedNode.geometry().hashCode());
			setBasket.remove(existedNode.geometry());
			this.nameToElement.remove(name);
		}
	}

	
}
