package cga;

import static org.junit.jupiter.api.Assertions.*;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import precision.PrecisionConfiguration;
import precision.model.IPrecisionConfiguration;
import spatial.algorithms.bestFittingPlane.BestFittingPlaneAlgorithm;
import spatial.geometry.Plane3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.bestFittingPlane.IBestFittingPlaneAlgorithm;
import spatial.model.algorithms.bestFittingPlane.IBestFittingPlaneResult;

class BestFittingPlaneFixture 
{
	private static IPrecisionConfiguration precisionConfiguration;
	
	@BeforeAll
	static void setUpBeforeClass() throws Exception 
	{
		precisionConfiguration = new PrecisionConfiguration(4, 0.0001);
	}

	@Test
	void testCalculate1()
	{
		//arrange
		
		//exact in plane points
		Vector3d[] points = new Vector3d[]
				{
					new Vector3d(3, 7, -6.25, precisionConfiguration),
					new Vector3d(-2, 4, -1.5, precisionConfiguration),
					new Vector3d(-4, 2, 1, precisionConfiguration),
					new Vector3d(1, 1, -0.75, precisionConfiguration)
				};

		//expected plane - plane(p, 2*x+3*y+4*z-2 = 0, [x, y, z])
		Vector3d expectedNormalVector = new Vector3d(0.3713906, 0.557086, 0.742781, precisionConfiguration);
		
		//act
		IBestFittingPlaneAlgorithm algorithm = new BestFittingPlaneAlgorithm(precisionConfiguration);
		IBestFittingPlaneResult result = algorithm.calculate(points);
		
		//assert algorithm has no errors
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		//assert distances to source points
		Plane3d plane = result.bestFittingPlane();
		for (int i = 0; i < points.length; i++)
		{
			Vector3d v = points[i].substract(plane.pointOnThePlane(), precisionConfiguration);
			double distance = plane.normalUnitVector().dotProduct(v);
			
			assertEquals(precisionConfiguration.compareWithAccuracy(0, distance), 0, "Not expected distance between plane and source point.");
		}
		
		//assert normal vector
		boolean normalVectorsEqalse = expectedNormalVector.equals(plane.normalUnitVector()) || expectedNormalVector.equals(plane.normalUnitVector().mul(-1, precisionConfiguration));
		assertTrue(normalVectorsEqalse, "Not expected normal vector.");
	}
	
	@Test
	void testCalculate2()
	{
		//arrange
		
		//exact in plane points
		Vector3d[] points = new Vector3d[]
				{
					//0.1
					new Vector3d(3, 7, -6.25, precisionConfiguration),
					//-0.09
					new Vector3d(-2, 4, -1.5, precisionConfiguration),
					//0.025
					new Vector3d(-4, 2, 1, precisionConfiguration),
					//-0.035
					new Vector3d(1, 1, -0.75, precisionConfiguration)
				};

		//expected plane - plane(p, 2*x+3*y+4*z-2 = 0, [x, y, z])
		double[] noisyDistances = new double[] {0.1, -0.09, 0.025, -0.035};
		Vector3d expectedNormalVector = new Vector3d(0.3713906, 0.557086, 0.742781, precisionConfiguration);
		double noisyDistancesSquaresSum = 0;
		for (int i = 0; i < noisyDistances.length; i++)
		{
			noisyDistancesSquaresSum += noisyDistances[i] * noisyDistances[i];
		}
		
		//noisy exact points
		for (int i = 0; i < points.length; i++)
		{
			points[i] = points[i].add(expectedNormalVector.mul(noisyDistances[i], precisionConfiguration), precisionConfiguration);
		}
		
		//act
		IBestFittingPlaneAlgorithm algorithm = new BestFittingPlaneAlgorithm(precisionConfiguration);
		IBestFittingPlaneResult result = algorithm.calculate(points);

		//assert algorithm has no errors
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		//assert sum of square distances should be less or equal to noisy distances to expected plane
		double sumOfDistancesSquares = 0;
		for (int i = 0; i < points.length; i++)
		{
			Vector3d v = points[i].substract(result.bestFittingPlane().pointOnThePlane(), precisionConfiguration);
			double distance = result.bestFittingPlane().normalUnitVector().dotProduct(v);
			sumOfDistancesSquares += distance * distance;
		}
		assertTrue(sumOfDistancesSquares <= noisyDistancesSquaresSum);

	}

}
