package cga;

import static org.junit.jupiter.api.Assertions.*;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import precision.PrecisionConfiguration;
import precision.model.IPrecisionConfiguration;
import spatial.algorithms.csTransform.CsTransformationAlgorithm;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.csTransformation.ICsTransformationAlgorithm;
import spatial.model.algorithms.csTransformation.ICsTransformationResult;

class CsTransformationFixture
{

	private static IPrecisionConfiguration precisionConfiguration;
	
	@BeforeAll
	static void setUpBeforeClass() throws Exception
	{
		precisionConfiguration = new PrecisionConfiguration(4, 0.0001);
	}

	@Test
	void testCalculate01()
	{
		//test of rotation without shifting local CS
		
		//arrange
		Vector3d normalPlaneOrt = new Vector3d(0.8004778, -.3121436, 0.5116658, precisionConfiguration);
		Vector3d inPlaneOrt = new Vector3d(0.5116656, 0.8004775, -.3121437, precisionConfiguration);
		Vector3d localCSOrigin = new Vector3d(0, 0, 0, precisionConfiguration);
		Vector3d[] globalCsPoints = new Vector3d[] { new Vector3d(0, 0, 0, precisionConfiguration), new Vector3d(3, 4, 5, precisionConfiguration) };
		
		double[] expectedRotationMatrix = new double[] {.5116656, -.3121435, .8004778, .8004775, .5116660, -.3121436, -.3121437, .8004777, .5116658};
		Vector3d[] expectedLocalCsPoints = new Vector3d[] { new Vector3d(0, 0, 0, precisionConfiguration), new Vector3d(4.2888118, 2.8873785, 4.8238087, precisionConfiguration) };
		
		//act
		ICsTransformationAlgorithm algorithm = new CsTransformationAlgorithm(precisionConfiguration);
		ICsTransformationResult result = algorithm.calculate(normalPlaneOrt, inPlaneOrt, localCSOrigin, globalCsPoints);
		
		//assert
		assertFalse(result.hasErrors(), "There are errors.");
		
		assertEquals(expectedLocalCsPoints[0], result.localPoints()[0], "Unexpected transformed point #0 coordinates.");
		assertEquals(expectedLocalCsPoints[1], result.localPoints()[1], "Unexpected transformed point #1 coordinates.");
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[0], result.rotationMatrix().get(0, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[1], result.rotationMatrix().get(0, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[2], result.rotationMatrix().get(0, 2)) == 0);
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[3], result.rotationMatrix().get(1, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[4], result.rotationMatrix().get(1, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[5], result.rotationMatrix().get(1, 2)) == 0);
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[6], result.rotationMatrix().get(2, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[7], result.rotationMatrix().get(2, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[8], result.rotationMatrix().get(2, 2)) == 0);
	}
	
	@Test
	void testCalculate02()
	{
		//test shifting local CS
		
		//arrange
		Vector3d normalPlaneOrt = new Vector3d(0, 0, 1, precisionConfiguration);
		Vector3d inPlaneOrt = new Vector3d(1, 0, 0, precisionConfiguration);
		Vector3d localCSOrigin = new Vector3d(2, 2, 2, precisionConfiguration);
		Vector3d[] globalCsPoints = new Vector3d[] { new Vector3d(2, 2, 2, precisionConfiguration) };
		
		double[] expectedRotationMatrix = new double[] {1, 0, 0, 0, 1, 0, 0, 0, 1};
		Vector3d[] expectedLocalCsPoints = new Vector3d[] { new Vector3d(0, 0, 0, precisionConfiguration) };
		
		//act
		ICsTransformationAlgorithm algorithm = new CsTransformationAlgorithm(precisionConfiguration);
		ICsTransformationResult result = algorithm.calculate(normalPlaneOrt, inPlaneOrt, localCSOrigin, globalCsPoints);
		
		//assert
		assertFalse(result.hasErrors(), "There are errors.");
		
		assertEquals(expectedLocalCsPoints[0], result.localPoints()[0], "Unexpected transformed point #0 coordinates.");
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[0], result.rotationMatrix().get(0, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[1], result.rotationMatrix().get(0, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[2], result.rotationMatrix().get(0, 2)) == 0);
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[3], result.rotationMatrix().get(1, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[4], result.rotationMatrix().get(1, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[5], result.rotationMatrix().get(1, 2)) == 0);
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[6], result.rotationMatrix().get(2, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[7], result.rotationMatrix().get(2, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[8], result.rotationMatrix().get(2, 2)) == 0);
	}
	
	@Test
	void testCalculate03()
	{
		//test shift and transform
		//assert length and angle of 3 points should stay the same after transformation
		
		
		//arrange
		Vector3d normalPlaneOrt = new Vector3d(0.8004778, -.3121436, 0.5116658, precisionConfiguration);
		Vector3d inPlaneOrt = new Vector3d(0.5116656, 0.8004775, -.3121437, precisionConfiguration);
		Vector3d localCSOrigin = new Vector3d(2, 3, 4, precisionConfiguration);
		
		Vector3d[] globalCsPoints = new Vector3d[] { new Vector3d(0, 0, 4, precisionConfiguration), new Vector3d(3, 4, 5, precisionConfiguration), new Vector3d(-2, 3, 0, precisionConfiguration) };
		
		double[] expectedRotationMatrix = new double[] {.5116656, -.3121435, .8004778, .8004775, .5116660, -.3121436, -.3121437, .8004777, .5116658};
		
		
		//act
		ICsTransformationAlgorithm algorithm = new CsTransformationAlgorithm(precisionConfiguration);
		ICsTransformationResult result = algorithm.calculate(normalPlaneOrt, inPlaneOrt, localCSOrigin, globalCsPoints);
		
		//assert algorithm completed without errors
		assertFalse(result.hasErrors(), "There are errors.");
		
		//assert rotation matrix
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[0], result.rotationMatrix().get(0, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[1], result.rotationMatrix().get(0, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[2], result.rotationMatrix().get(0, 2)) == 0);
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[3], result.rotationMatrix().get(1, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[4], result.rotationMatrix().get(1, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[5], result.rotationMatrix().get(1, 2)) == 0);
		
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[6], result.rotationMatrix().get(2, 0)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[7], result.rotationMatrix().get(2, 1)) == 0);
		assertTrue(precisionConfiguration.compareWithAccuracy(expectedRotationMatrix[8], result.rotationMatrix().get(2, 2)) == 0);
		
		//assert the length for vectors built from 3 points stay the same
		for (int i = 0; i < globalCsPoints.length; i++)
		{
			Vector3d g1 = globalCsPoints[i % globalCsPoints.length];
			Vector3d g2 = globalCsPoints[(i + 1) % globalCsPoints.length];
			
			Vector3d l1 = result.localPoints()[i % globalCsPoints.length];
			Vector3d l2 = result.localPoints()[(i + 1) % globalCsPoints.length];
			
			double gLenght = g2.substract(g1, precisionConfiguration).euqlideanNorm();
			double lLenght = l2.substract(l1, precisionConfiguration).euqlideanNorm();
			
			assertEquals(0, precisionConfiguration.compareWithAccuracy(gLenght, lLenght), "Not equal lengths of vectors after transformation.");
		}
		
		//assert the angles for vectors built from 3 points stay the same
		for (int i = 0; i < globalCsPoints.length; i++)
		{
			Vector3d g1 = globalCsPoints[i % globalCsPoints.length];
			Vector3d g2 = globalCsPoints[(i + 1) % globalCsPoints.length];
			Vector3d g3 = globalCsPoints[(i + 2) % globalCsPoints.length];
			
			Vector3d l1 = result.localPoints()[i % globalCsPoints.length];
			Vector3d l2 = result.localPoints()[(i + 1) % globalCsPoints.length];
			Vector3d l3 = result.localPoints()[(i + 2) % globalCsPoints.length];
			
			Vector3d g12 = g2.substract(g1, precisionConfiguration);
			Vector3d g13 = g3.substract(g1, precisionConfiguration);
			
			Vector3d l12 = l2.substract(l1, precisionConfiguration);
			Vector3d l13 = l3.substract(l1, precisionConfiguration);
			
			double ga = g12.dotProduct(g13) / (g12.euqlideanNorm() * g13.euqlideanNorm());
			double la = l12.dotProduct(l13) / (l12.euqlideanNorm() * l13.euqlideanNorm());
			
			assertEquals(0, precisionConfiguration.compareWithAccuracy(ga, la), "Not equal angles of vectors after transformation.");
		}
	}

}
