package cga;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertFalse;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import precision.PrecisionConfiguration;
import precision.model.IPrecisionConfiguration;
import spatial.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationAlgorithm;
import spatial.geometry.Edge2d;
import spatial.geometry.Vector2d;
import spatial.model.algorithms.edgeToEdgeLocation.EdgeToEdgeLocationKind;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationAlgorithm;
import spatial.model.algorithms.edgeToEdgeLocation.IEdgeToEdgeLocationResult;

class EdgeToEdgeLocationFixture
{

	private static IPrecisionConfiguration precisionConfiguration;
	
	@BeforeAll
	static void setUpBeforeClass() throws Exception
	{
		precisionConfiguration = new PrecisionConfiguration(4, 0.0001);
	}

	@Test
	void testCalculate01()
	{
		//test intersection in one interior point for both edges 
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(5, 20, precisionConfiguration), new Vector2d(35, 10, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(10, 10, precisionConfiguration), new Vector2d(40, 25, precisionConfiguration));
		
		Vector2d expectedCrossPoint = new Vector2d(20, 15, precisionConfiguration);
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.NotParallel, result.locationCode(), "Not expected location kind.");
		assertEquals(expectedCrossPoint, result.intersectionPoint(), "Not expected point of intersection.");
		
		assertEquals(0, precisionConfiguration.compareWithAccuracy(result.r0(), 0.5), "Not expected r0.");
		assertEquals(0, precisionConfiguration.compareWithAccuracy(result.t0(), 0.3333333), "Not expected t0.");
		
		assertTrue("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
	
	@Test
	void testCalculate02()
	{
		//test intersection in one end point for one of the edges 
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(5, 20, precisionConfiguration), new Vector2d(35, 10, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(30, 5, precisionConfiguration), new Vector2d(40, 15, precisionConfiguration));
		
		Vector2d expectedCrossPoint = new Vector2d(35, 10, precisionConfiguration);
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.NotParallel, result.locationCode(), "Not expected location kind.");
		assertEquals(expectedCrossPoint, result.intersectionPoint(), "Not expected point of intersection.");
		
		assertEquals(0, precisionConfiguration.compareWithAccuracy(result.r0(), 1), "Not expected r0.");
		assertEquals(0, precisionConfiguration.compareWithAccuracy(result.t0(), 0.5), "Not expected t0.");
		
		assertTrue("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
	
	@Test
	void testCalculate03()
	{
		//test parallel edges
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(10, 10, precisionConfiguration), new Vector2d(30, 20, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(40, 20, precisionConfiguration), new Vector2d(50, 25, precisionConfiguration));
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.Parallel, result.locationCode(), "Not expected location kind.");
		
		assertFalse("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
	
	@Test
	void testCalculate04()
	{
		//test collinear not overlap
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(10, 10, precisionConfiguration), new Vector2d(30, 20, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(40, 25, precisionConfiguration), new Vector2d(50, 30, precisionConfiguration));
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.Collinear, result.locationCode(), "Not expected location kind.");
		
		assertTrue(result.rc() > 1);
		assertTrue(result.rd() > 1);
		
		assertFalse("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
	
	@Test
	void testCalculate05()
	{
		//test collinear overlap at node
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(10, 10, precisionConfiguration), new Vector2d(30, 20, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(30, 20, precisionConfiguration), new Vector2d(50, 30, precisionConfiguration));
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.OverlapAtNode, result.locationCode(), "Not expected location kind.");
		
		assertTrue(precisionConfiguration.compareWithAccuracy(1, result.rc()) == 0);
		assertTrue(result.rd() > 1);
		
		assertFalse("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertTrue("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
	
	@Test
	void testCalculate06()
	{
		//test collinear overlap at more than one point
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(10, 10, precisionConfiguration), new Vector2d(30, 20, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(20, 15, precisionConfiguration), new Vector2d(50, 30, precisionConfiguration));
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.Overlap, result.locationCode(), "Not expected location kind.");
		
		assertTrue(result.rc() > 0 && result.rc() < 1);
		assertTrue(result.rd() > 1);
		
		assertFalse("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}

	@Test
	void testCalculate07()
	{
		//test not intersected edges, but intersected at node lines
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(15, 35, precisionConfiguration), new Vector2d(0, 20, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(40, 10, precisionConfiguration), new Vector2d(30, 20, precisionConfiguration));
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.NotParallel, result.locationCode(), "Not expected location kind.");
			
		assertFalse("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
	
	@Test
	void testCalculate08()
	{
		//one edge touch other in inner node
		
		//arrange
		Edge2d edge1 = new Edge2d(new Vector2d(30, 20, precisionConfiguration), new Vector2d(15, 30, precisionConfiguration));
		Edge2d edge2 = new Edge2d(new Vector2d(50, 15, precisionConfiguration), new Vector2d(40, 0, precisionConfiguration));
		
		//act
		IEdgeToEdgeLocationAlgorithm algorithm = new EdgeToEdgeLocationAlgorithm(precisionConfiguration);
		IEdgeToEdgeLocationResult result = algorithm.calculate(edge1, edge2);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(EdgeToEdgeLocationKind.NotParallel, result.locationCode(), "Not expected location kind.");
		
		assertFalse("Not expected mark edgesIntersectInside", result.edgesIntersectInside());
		assertFalse("Not expected mark edgesIntersectAtNode", result.edgesIntersectAtNode());
	}
}
