package cga;

import static org.junit.jupiter.api.Assertions.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import precision.PreciseDoubleFactory;
import precision.model.IPreciseDoubleFactory;
import spatial.NodeFactory;
import spatial.SpatialModel;
import spatial.geometry.Vector3d;
import spatial.model.IElement;
import spatial.model.ISpatialModel;
import spatial.model.elements.INode;
import spatial.model.elements.set.INodeSet;
import spatial.model.elements.set.exception.ElementWithTheSameGeometryAlreadyExistsException;
import spatial.model.elements.set.exception.ElementWithTheSameNameAlreadyExistsException;

class NodeSetFixture 
{
	private ISpatialModel spatialModel;
	private IPreciseDoubleFactory preciseDoubleFactory;
	
	@BeforeEach
	void setUpBeforeEach() throws Exception
	{
		this.spatialModel = new SpatialModel();
		this.preciseDoubleFactory = new PreciseDoubleFactory(this.spatialModel.precisionConfiguration());
	}

	@Test
	void testEnsureNotThrows() 
	{
		INodeSet nodes = this.spatialModel.nodes();

		INode nodeA = (new NodeFactory()).create("nodeA", new Vector3d(10, 10, 10, this.preciseDoubleFactory));
		INode nodeB = (new NodeFactory()).create("nodeB", new Vector3d(20, 20, 20, this.preciseDoubleFactory));
		try 
		{
			INode ensuredNodeA = (INode)nodes.ensure(nodeA);
			INode ensuredNodeB = (INode)nodes.ensure(nodeB);
			
			assertEquals(nodeA, ensuredNodeA);
			assertEquals(nodeB, ensuredNodeB);
		}
		catch (Throwable throwable) 
		{
			fail(throwable);
		}
	}
	
	@Test
	void testEnsureDuplicateName() 
	{
		ISpatialModel spatialModel = new SpatialModel();
		INodeSet nodes = spatialModel.nodes();
		
		INode nodeA = (new NodeFactory()).create("nodeA", new Vector3d(10, 20, 30, this.preciseDoubleFactory));
		INode nodeB = (new NodeFactory()).create("nodeA", new Vector3d(41, 42, 43, this.preciseDoubleFactory));
		
		assertThrows(ElementWithTheSameNameAlreadyExistsException.class, () -> {
			nodes.ensure(nodeA);
			nodes.ensure(nodeB);
		});
	}
	
	@Test
	void testEnsureDuplicateGeometry() 
	{
		ISpatialModel spatialModel = new SpatialModel();
		INodeSet nodes = spatialModel.nodes();
		
		INode nodeA = (new NodeFactory()).create("nodeA", new Vector3d(41, 42, 43, this.preciseDoubleFactory));
		INode nodeB = (new NodeFactory()).create("nodeB", new Vector3d(41, 42, 43, this.preciseDoubleFactory));
		
		assertThrows(ElementWithTheSameGeometryAlreadyExistsException.class, () -> {
			nodes.ensure(nodeA);
			nodes.ensure(nodeB);
		});
	}

	@Test
	void testContains() 
	{
		ISpatialModel spatialModel = new SpatialModel();
		INodeSet nodes = spatialModel.nodes();
		
		INode nodeA = (new NodeFactory()).create("nodeA", new Vector3d(41, 42, 43, this.preciseDoubleFactory));
		INode nodeB = (new NodeFactory()).create("nodeB", new Vector3d(43, 42, 41, this.preciseDoubleFactory));

		try 
		{	
			assertFalse(nodes.contains("nodeC"));
			
			nodes.ensure(nodeA);
			nodes.ensure(nodeB);
			
			assertTrue(nodes.contains("nodeA"));
			assertTrue(nodes.contains("nodeB"));
			assertFalse(nodes.contains("nodeC"));
		} 
		catch (Throwable throwable) 
		{
			fail(throwable);
		}
	}

	@Test
	void testGet() 
	{
		ISpatialModel spatialModel = new SpatialModel();
		INodeSet nodes = spatialModel.nodes();
		
		INode nodeA = (new NodeFactory()).create("nodeA", new Vector3d(41, 42, 43, this.preciseDoubleFactory));
		INode nodeB = (new NodeFactory()).create("nodeB", new Vector3d(43, 42, 41, this.preciseDoubleFactory));

		try 
		{	
			//on empty, initial collection
			IElement noNodeByName = nodes.get("nodeC");
			IElement noNodeByGeometry = nodes.get("nodeC");
			
			assertNull(noNodeByName);
			assertNull(noNodeByGeometry);
			
			//on warmed collection
			nodes.ensure(nodeA);
			nodes.ensure(nodeB);
			
			IElement nodeAByName = nodes.get("nodeA");
			IElement nodeBByName = nodes.get("nodeB");
			noNodeByName = nodes.get("nodeC");
			
			IElement nodeAByGeometry = nodes.get(new Vector3d(41, 42, 43, this.preciseDoubleFactory));
			IElement nodeBByGeometry = nodes.get(new Vector3d(43, 42, 41, this.preciseDoubleFactory));
			noNodeByGeometry = nodes.get(new Vector3d(43, 43, 43, this.preciseDoubleFactory));
			
			assertNotNull(nodeAByName);
			assertNotNull(nodeBByName);
			assertNull(noNodeByName);
			
			assertNotNull(nodeAByGeometry);
			assertNotNull(nodeBByGeometry);
			assertNull(noNodeByGeometry);
			
			assertEquals(nodeAByName, nodeAByGeometry);
			assertEquals(nodeBByName, nodeBByGeometry);
		} 
		catch (Throwable throwable) 
		{
			fail(throwable);
		}
	}

	@Test
	void testRemove() 
	{
		ISpatialModel spatialModel = new SpatialModel();
		INodeSet nodes = spatialModel.nodes();
		
		INode nodeA = (new NodeFactory()).create("nodeA", new Vector3d(41, 42, 43, this.preciseDoubleFactory));
		INode nodeB = (new NodeFactory()).create("nodeB", new Vector3d(43, 42, 41, this.preciseDoubleFactory));

		try 
		{
			nodes.ensure(nodeA);
			nodes.ensure(nodeB);
			
			nodes.remove("nodeA");

			IElement nodeAByName = nodes.get("nodeA");
			IElement nodeBByName = nodes.get("nodeB");

			assertNull(nodeAByName);
			assertNotNull(nodeBByName);

			nodes.remove("nodeB");
			
			nodeAByName = nodes.get("nodeA");
			nodeBByName = nodes.get("nodeB");

			assertNull(nodeAByName);
			assertNull(nodeBByName);
		} 
		catch (Throwable throwable) 
		{
			fail(throwable);
		}
	}

}
