package cga;

import static org.junit.jupiter.api.Assertions.*;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import precision.PrecisionConfiguration;
import precision.model.IPrecisionConfiguration;
import spatial.algorithms.pointToEdgeDistance.PointToEdgeDistanceAlgorithm;
import spatial.geometry.Edge3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.pointToEdgeDistance.IPointToEdgeDistanceAlgorithm;
import spatial.model.algorithms.pointToEdgeDistance.IPointToEdgeDistanceResult;
import spatial.model.algorithms.pointToEdgeDistance.PointToEdgeLocationKind;

class PointToEdgeDistanceFixture
{

	private static IPrecisionConfiguration precisionConfiguration;
	
	@BeforeAll
	static void setUpBeforeClass() throws Exception 
	{
		precisionConfiguration = new PrecisionConfiguration(4, 0.0001);
	}
		
	@Test
	void testCalculate01()
	{
		// ---input---
		// edge: (15, 0, 10) - (35, 0, 10)
		// point: (45, 0, 30)
		
		// ---expected output---
		// distance: 22.36067
		// distance kind: EndPoint
		// projection point on the edge line: (45, 0, 10)
		
		//arrange
		Edge3d edge = new Edge3d(new Vector3d(15, 0, 10, precisionConfiguration), new Vector3d(35, 0, 10, precisionConfiguration));
		Vector3d point = new Vector3d(45, 0, 30, precisionConfiguration);
		
		//act
		IPointToEdgeDistanceAlgorithm algorithm = new PointToEdgeDistanceAlgorithm(precisionConfiguration);
		IPointToEdgeDistanceResult result = algorithm.calculate(edge, point);
		
		//assert
		assertFalse(result.hasErrors());
		assertEquals(0, precisionConfiguration.compareWithAccuracy(22.3606, result.distance()));
		assertEquals(PointToEdgeLocationKind.EndPoint, result.distancePointLocationKind());
		assertEquals(edge.vectorB(), result.edgeDistancePoint(), "projection point doesn't match expected value");
	}
	
	@Test
	void testCalculate02()
	{
		// data is the same as for testCalculate01 but rotated by matrix [[0.5116656,-0.3121435,0.8004778],[0.8004775,0.5116660,-0.3121436],[-0.3121437,0.8004777,0.5116658]]
	
		// ---expected output---
		// distance: 22.36067
		// distance kind: EndPoint
		// projection point on the edge line: (31.02973, 32.9000515, -8.9298085)
		
		//arrange
		
		Edge3d edge = new Edge3d(new Vector3d(15.6797646348022, 8.88572931678020, 0.434506048417201, precisionConfiguration), 
				new Vector3d(25.9130823497966, 24.8952820138922, -5.80836436368960, precisionConfiguration));
		Vector3d point = new Vector3d(47.0392939044066, 26.6571879503406, 1.30351814525160, precisionConfiguration);
		
		//act
		IPointToEdgeDistanceAlgorithm algorithm = new PointToEdgeDistanceAlgorithm(precisionConfiguration);
		IPointToEdgeDistanceResult result = algorithm.calculate(edge, point);
		
		//assert
		assertFalse(result.hasErrors());
		assertEquals(0, precisionConfiguration.compareWithAccuracy(22.3606, result.distance()));
		assertEquals(PointToEdgeLocationKind.EndPoint, result.distancePointLocationKind());
		assertEquals(edge.vectorB(), result.edgeDistancePoint(), "projection point doesn't match expected value");
	}
	
	@Test
	void testCalculate03()
	{
		//arrange
		Edge3d edge = new Edge3d(new Vector3d(0, 0, 0, precisionConfiguration), new Vector3d(20, 0, 0, precisionConfiguration));
		Vector3d point = new Vector3d(12.5, 7.5, 0, precisionConfiguration);
		
		//act
		IPointToEdgeDistanceAlgorithm algorithm = new PointToEdgeDistanceAlgorithm(precisionConfiguration);
		IPointToEdgeDistanceResult result = algorithm.calculate(edge, point);
		
		//assert
		assertFalse(result.hasErrors());
		assertEquals(0, precisionConfiguration.compareWithAccuracy(7.5, result.distance()));
		assertEquals(PointToEdgeLocationKind.InnerPoint, result.distancePointLocationKind());
		assertEquals(new Vector3d(12.5000, 0, 0, precisionConfiguration), result.edgeDistancePoint(), "projection point doesn't match expected value");
	}

	@Test
	void testCalculate04()
	{
		//arrange
		Edge3d edge = new Edge3d(new Vector3d(30, 10, 0, precisionConfiguration), new Vector3d(37.5, 25, 0, precisionConfiguration));
		Vector3d point = new Vector3d(22.5, 17.5, 0, precisionConfiguration);
		
		//act
		IPointToEdgeDistanceAlgorithm algorithm = new PointToEdgeDistanceAlgorithm(precisionConfiguration);
		IPointToEdgeDistanceResult result = algorithm.calculate(edge, point);
		
		//assert
		assertFalse(result.hasErrors());
		assertEquals(0, precisionConfiguration.compareWithAccuracy(10.0623, result.distance()));
		assertEquals(PointToEdgeLocationKind.InnerPoint, result.distancePointLocationKind());
		assertEquals(new Vector3d(31.5000, 13, 0, precisionConfiguration), result.edgeDistancePoint(), "projection point doesn't match expected value");
	}
}
