package cga;

import static org.junit.jupiter.api.Assertions.*;

import java.util.stream.IntStream;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import precision.PrecisionConfiguration;
import precision.model.IPrecisionConfiguration;
import spatial.algorithms.polygon2dSelfIntersection.Polygon2dSelfIntersectionAlgorithm;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionAlgorithm;
import spatial.model.algorithms.polygon2dSelfIntersection.IPolygon2dSelfIntersectionResult;
import spatial.model.algorithms.polygon2dSelfIntersection.Polygon2dSelfIntersectionResultCode;

class Polygon2dSelfIntersectionFixture
{
	private static IPrecisionConfiguration precisionConfiguration;
	
	@BeforeAll
	static void setUpBeforeClass() throws Exception
	{
		precisionConfiguration = new PrecisionConfiguration(4, 0.0001);
	}

	@Test
	void testCalculate01()
	{
		//triangle
		
		//arrange
		Polygon3d polygon = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(10, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		//act
		IPolygon2dSelfIntersectionAlgorithm algorithm = new Polygon2dSelfIntersectionAlgorithm(precisionConfiguration);
		IPolygon2dSelfIntersectionResult result = algorithm.calculate(planeNormalVector, polygon);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(Polygon2dSelfIntersectionResultCode.NotSelfIntersected, result.resultCode(), "Not expected result code.");
	}
	
	@Test
	void testCalculate02()
	{
		//6-angle not intersected polygon
		
		//arrange
		Polygon3d polygon = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 5, 0, precisionConfiguration),
				new Vector3d(40, 5, 0, precisionConfiguration),
				new Vector3d(50, 15, 0, precisionConfiguration),
				new Vector3d(40, 25, 0, precisionConfiguration),
				new Vector3d(25, 40, 0, precisionConfiguration),
				new Vector3d(10, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		//act
		IPolygon2dSelfIntersectionAlgorithm algorithm = new Polygon2dSelfIntersectionAlgorithm(precisionConfiguration);
		IPolygon2dSelfIntersectionResult result = algorithm.calculate(planeNormalVector, polygon);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(Polygon2dSelfIntersectionResultCode.NotSelfIntersected, result.resultCode(), "Not expected result code.");
	}

	@Test
	void testCalculate03()
	{
		//6-angle, one of the nodes belongs touches the interior of the edge
		
		//arrange
		Polygon3d polygon = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 5, 0, precisionConfiguration),
				new Vector3d(40, 5, 0, precisionConfiguration),
				new Vector3d(50, 15, 0, precisionConfiguration),
				new Vector3d(40, 25, 0, precisionConfiguration),
				new Vector3d(25, 5, 0, precisionConfiguration),
				new Vector3d(10, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);		
		int expectedEdgeIndex1 = 0;
		int expectedEdgeIndex2 = 4;
		int expectedEdgeIndex3 = 3;
		
		//act
		IPolygon2dSelfIntersectionAlgorithm algorithm = new Polygon2dSelfIntersectionAlgorithm(precisionConfiguration);
		IPolygon2dSelfIntersectionResult result = algorithm.calculate(planeNormalVector, polygon);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(Polygon2dSelfIntersectionResultCode.SelfIntersected, result.resultCode(), "Not expected result code.");
		
		assertTrue((result.intersectedPair().EdgeIx1 == expectedEdgeIndex1 && result.intersectedPair().EdgeIx2 == expectedEdgeIndex2 || result.intersectedPair().EdgeIx2 == expectedEdgeIndex3) ||
				(result.intersectedPair().EdgeIx2 == expectedEdgeIndex1 && result.intersectedPair().EdgeIx1 == expectedEdgeIndex2 || result.intersectedPair().EdgeIx1 == expectedEdgeIndex3));
	}
	
	@Test
	void testCalculate04()
	{
		//6-angle, self-intersection at node
		
		//arrange
		Polygon3d polygon = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 5, 0, precisionConfiguration),
				new Vector3d(40, 5, 0, precisionConfiguration),
				new Vector3d(50, 15, 0, precisionConfiguration),
				new Vector3d(40, 25, 0, precisionConfiguration),
				new Vector3d(40, 5, 0, precisionConfiguration),
				new Vector3d(10, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);		
//		int[] expectedEdgeIndex1 = new int[] {0, 1};
//		int[] expectedEdgeIndex2 = new int[] {3, 4};
		
		//act
		IPolygon2dSelfIntersectionAlgorithm algorithm = new Polygon2dSelfIntersectionAlgorithm(precisionConfiguration);
		IPolygon2dSelfIntersectionResult result = algorithm.calculate(planeNormalVector, polygon);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(Polygon2dSelfIntersectionResultCode.SelfIntersected, result.resultCode(), "Not expected result code.");
		
//		boolean xCheck = IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedPair().EdgeIx1) && IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedPair().EdgeIx2) ||
//				IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedPair().EdgeIx1) && IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedPair().EdgeIx2);
//
//		assertTrue(xCheck);
	}
	
	@Test
	void testCalculate05()
	{
		//6-angle, self-intersection
		
		//arrange
		Polygon3d polygon = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 5, 0, precisionConfiguration),
				new Vector3d(40, 5, 0, precisionConfiguration),
				new Vector3d(50, 15, 0, precisionConfiguration),
				new Vector3d(40, 25, 0, precisionConfiguration),
				new Vector3d(25, 0, 0, precisionConfiguration),
				new Vector3d(10, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);		
		int[] expectedEdgeIndex1 = new int[] {0, 1};
		int[] expectedEdgeIndex2 = new int[] {3, 4};
		
		//act
		IPolygon2dSelfIntersectionAlgorithm algorithm = new Polygon2dSelfIntersectionAlgorithm(precisionConfiguration);
		IPolygon2dSelfIntersectionResult result = algorithm.calculate(planeNormalVector, polygon);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(Polygon2dSelfIntersectionResultCode.SelfIntersected, result.resultCode(), "Not expected result code.");
		
		boolean xCheck = IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedPair().EdgeIx1) && IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedPair().EdgeIx2) ||
				IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedPair().EdgeIx1) && IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedPair().EdgeIx2);

		assertTrue(xCheck);
	}
}
