package cga;

import static org.junit.jupiter.api.Assertions.*;

import java.util.stream.IntStream;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import precision.PrecisionConfiguration;
import precision.model.IPrecisionConfiguration;
import spatial.algorithms.polygonToPolygonLocation.PolygonToPolygonLocationAlgorithm;
import spatial.geometry.Polygon3d;
import spatial.geometry.Vector3d;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationAlgorithm;
import spatial.model.algorithms.polygonToPolygonLocation.IPolygonToPolygonLocationResult;
import spatial.model.algorithms.polygonToPolygonLocation.PolygonToPolygonLocationKind;

class PolygonToPolygonLocationFixture
{
	private static IPrecisionConfiguration precisionConfiguration;
	
	@BeforeAll
	static void setUpBeforeClass() throws Exception
	{
		precisionConfiguration = new PrecisionConfiguration(4, 0.0001);
	}

	@Test
	void testCalculate01()
	{
		//two not intersected polygons
		
		//arrange
		Polygon3d polygonA = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration),
				new Vector3d(25, 40, 0, precisionConfiguration),
				new Vector3d(10, 30, 0, precisionConfiguration)
		});
		
		Polygon3d polygonB = new Polygon3d(new Vector3d[] {
				new Vector3d(50, 10, 0, precisionConfiguration),
				new Vector3d(70, 10, 0, precisionConfiguration),
				new Vector3d(60, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		//act
		IPolygonToPolygonLocationAlgorithm algorithm = new PolygonToPolygonLocationAlgorithm(precisionConfiguration);
		IPolygonToPolygonLocationResult result = algorithm.calculate(planeNormalVector, polygonA, polygonB);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(PolygonToPolygonLocationKind.NotIntersects, result.locationCode(), "Not expected result code.");
	}

	@Test
	void testCalculate02()
	{
		//two not intersected polygons, one with self intersection
		
		//arrange
		Polygon3d polygonA = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration),
				new Vector3d(25, 10, 0, precisionConfiguration),
				new Vector3d(10, 30, 0, precisionConfiguration)
		});
		
		Polygon3d polygonB = new Polygon3d(new Vector3d[] {
				new Vector3d(50, 10, 0, precisionConfiguration),
				new Vector3d(70, 10, 0, precisionConfiguration),
				new Vector3d(60, 25, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		//act
		IPolygonToPolygonLocationAlgorithm algorithm = new PolygonToPolygonLocationAlgorithm(precisionConfiguration);
		IPolygonToPolygonLocationResult result = algorithm.calculate(planeNormalVector, polygonA, polygonB);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(PolygonToPolygonLocationKind.NotIntersects, result.locationCode(), "Not expected result code.");
	}
	
	@Test
	void testCalculate03()
	{
		//two intersected at node polygons
		
		//arrange
		Polygon3d polygonA = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration),
				new Vector3d(25, 40, 0, precisionConfiguration),
				new Vector3d(10, 30, 0, precisionConfiguration)
		});
		
		Polygon3d polygonB = new Polygon3d(new Vector3d[] {
				new Vector3d(40, 10, 0, precisionConfiguration),
				new Vector3d(50, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		int[] expectedEdgeIndex1 = new int[] {1, 2};
		int[] expectedEdgeIndex2 = new int[] {1, 2};
		
		//act
		IPolygonToPolygonLocationAlgorithm algorithm = new PolygonToPolygonLocationAlgorithm(precisionConfiguration);
		IPolygonToPolygonLocationResult result = algorithm.calculate(planeNormalVector, polygonA, polygonB);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(PolygonToPolygonLocationKind.Intersects, result.locationCode(), "Not expected result code.");
		
		boolean xCheck = IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2) ||
				IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2);

		assertTrue(xCheck);
	}
	
	@Test
	void testCalculate04()
	{
		//two intersected polygons (node + inside)
		
		//arrange
		Polygon3d polygonA = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration),
				new Vector3d(25, 40, 0, precisionConfiguration),
				new Vector3d(10, 30, 0, precisionConfiguration)
		});
		
		Polygon3d polygonB = new Polygon3d(new Vector3d[] {
				new Vector3d(40, 10, 0, precisionConfiguration),
				new Vector3d(50, 10, 0, precisionConfiguration),
				new Vector3d(40, 40, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		int[] expectedEdgeIndex1 = new int[] {1, 2};
		int[] expectedEdgeIndex2 = new int[] {1, 2};
		
		//act
		IPolygonToPolygonLocationAlgorithm algorithm = new PolygonToPolygonLocationAlgorithm(precisionConfiguration);
		IPolygonToPolygonLocationResult result = algorithm.calculate(planeNormalVector, polygonA, polygonB);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(PolygonToPolygonLocationKind.Intersects, result.locationCode(), "Not expected result code.");
		
		boolean xCheck = IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2) ||
				IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2);

		assertTrue(xCheck);
	}
	
	@Test
	void testCalculate05()
	{
		//two intersected polygons (at edge)
		
		//arrange
		Polygon3d polygonA = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration),
				new Vector3d(25, 40, 0, precisionConfiguration),
				new Vector3d(10, 30, 0, precisionConfiguration)
		});
		
		Polygon3d polygonB = new Polygon3d(new Vector3d[] {
				new Vector3d(25, 0, 0, precisionConfiguration),
				new Vector3d(50, 10, 0, precisionConfiguration),
				new Vector3d(45, 40, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		int[] expectedEdgeIndex1 = new int[] {0, 1, 2};
		int[] expectedEdgeIndex2 = new int[] { 2 };
		
		//act
		IPolygonToPolygonLocationAlgorithm algorithm = new PolygonToPolygonLocationAlgorithm(precisionConfiguration);
		IPolygonToPolygonLocationResult result = algorithm.calculate(planeNormalVector, polygonA, polygonB);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(PolygonToPolygonLocationKind.Intersects, result.locationCode(), "Not expected result code.");
		
		boolean xCheck = IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2) ||
				IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2);

		assertTrue(xCheck);
	}
	
	@Test
	void testCalculate06()
	{
		//two intersected polygons
		
		//arrange
		Polygon3d polygonA = new Polygon3d(new Vector3d[] {
				new Vector3d(10, 10, 0, precisionConfiguration),
				new Vector3d(30, 10, 0, precisionConfiguration),
				new Vector3d(40, 30, 0, precisionConfiguration),
				new Vector3d(25, 40, 0, precisionConfiguration),
				new Vector3d(10, 30, 0, precisionConfiguration)
		});
		
		Polygon3d polygonB = new Polygon3d(new Vector3d[] {
				new Vector3d(25, 0, 0, precisionConfiguration),
				new Vector3d(50, 10, 0, precisionConfiguration),
				new Vector3d(35, 45, 0, precisionConfiguration)
		});
		Vector3d planeNormalVector = new Vector3d(0, 0, 1, precisionConfiguration);
		
		int[] expectedEdgeIndex1 = new int[] {0, 2};
		int[] expectedEdgeIndex2 = new int[] { 2 };
		
		//act
		IPolygonToPolygonLocationAlgorithm algorithm = new PolygonToPolygonLocationAlgorithm(precisionConfiguration);
		IPolygonToPolygonLocationResult result = algorithm.calculate(planeNormalVector, polygonA, polygonB);
		
		//assert
		assertFalse(result.hasErrors(), "Algorithm resulted with errors.");
		
		assertEquals(PolygonToPolygonLocationKind.Intersects, result.locationCode(), "Not expected result code.");
		
		boolean xCheck = IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2) ||
				IntStream.of(expectedEdgeIndex2).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx1) && IntStream.of(expectedEdgeIndex1).anyMatch(x -> x == result.intersectedEdgesPair().EdgeIx2);

		assertTrue(xCheck);
	}
}
