package cga;

import static org.junit.jupiter.api.Assertions.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import precision.PreciseDoubleFactory;
import precision.PrecisionConfiguration;
import precision.model.IPreciseDoubleFactory;
import precision.model.IPrecisionConfiguration;
import precision.model.PreciseDouble;

class PreciseDoubleFixture 
{
	private IPrecisionConfiguration precisionConfiguration;
	
	private IPreciseDoubleFactory preciseDoubleFactory;
	
	@BeforeEach
	void setUpBeforeEach() throws Exception
	{
		this.precisionConfiguration = new PrecisionConfiguration(4, 0.000009);
		this.preciseDoubleFactory = new PreciseDoubleFactory(this.precisionConfiguration);
    }
	
	@Test
	void testConfiguration()
	{
		assertEquals(4, this.precisionConfiguration.accuracy());
		assertEquals(10000, this.precisionConfiguration.iEpsilon());
	}
	
	@Test
	void testWholePart()
	{
		assertEquals(123, this.preciseDoubleFactory.create(123.123).wholePart());
		
		assertEquals(123, this.preciseDoubleFactory.create(123).wholePart());
		
		assertEquals(0, this.preciseDoubleFactory.create(0.123).wholePart());
		
		assertEquals(0, this.preciseDoubleFactory.create(-0.123).wholePart());
		
		assertEquals(0, this.preciseDoubleFactory.create(0).wholePart());
		
		assertEquals(123, this.preciseDoubleFactory.create(-123.123).wholePart());
		
		assertEquals(124, this.preciseDoubleFactory.create(-123.99999).wholePart());
		
		assertEquals(123, this.preciseDoubleFactory.create(-123.999).wholePart());
	}

	@Test
	void testDecimalPart()
	{
		assertEquals(1230, this.preciseDoubleFactory.create(123.123).decimalPart());
		
		assertEquals(0, this.preciseDoubleFactory.create(123).decimalPart());
		
		assertEquals(1230, this.preciseDoubleFactory.create(0.123).decimalPart());
		
		assertEquals(1230, this.preciseDoubleFactory.create(-0.123).decimalPart());
		
		assertEquals(0, this.preciseDoubleFactory.create(0).decimalPart());
		
		assertEquals(1235, this.preciseDoubleFactory.create(123.123456).decimalPart());
		
		assertEquals(1230, this.preciseDoubleFactory.create(-123.123).decimalPart());
		
		assertEquals(0, this.preciseDoubleFactory.create(-123.99999).decimalPart());
		
		assertEquals(9999, this.preciseDoubleFactory.create(-123.9999).decimalPart());
	}
	
	@Test
	void testSign()
	{
		assertEquals(1, this.preciseDoubleFactory.create(123.123).sign());
		assertEquals(1, this.preciseDoubleFactory.create(0.123).sign());
		assertEquals(-1, this.preciseDoubleFactory.create(-0.123).sign());
		assertEquals(-1, this.preciseDoubleFactory.create(-123.123).sign());
		
		
		assertEquals(0, this.preciseDoubleFactory.create(-0.00001).sign());
		assertEquals(-1, this.preciseDoubleFactory.create(-0.00009).sign());
	}
	
	@Test
	void testEquals1() 
	{
		assertNotEquals(this.preciseDoubleFactory.create(123.123456), this.preciseDoubleFactory.create(123.1234));
		assertEquals(this.preciseDoubleFactory.create(123.123445), this.preciseDoubleFactory.create(123.1234));
		assertNotEquals(this.preciseDoubleFactory.create(123.123), this.preciseDoubleFactory.create(123.1234));
		
		assertNotEquals(this.preciseDoubleFactory.create(-123.123456), this.preciseDoubleFactory.create(-123.1234));
		assertEquals(this.preciseDoubleFactory.create(-123.123444), this.preciseDoubleFactory.create(-123.1234));
		assertNotEquals(this.preciseDoubleFactory.create(-123.123), this.preciseDoubleFactory.create(-123.1234));
	}
	
	@Test
	void testEquals2() 
	{
		assertNotEquals(this.preciseDoubleFactory.create(0.123456), this.preciseDoubleFactory.create(-0.1234));
		
		assertNotEquals(this.preciseDoubleFactory.create(-0.123456), this.preciseDoubleFactory.create(-0.123));
	}
	
	@Test
	void testEquals3()
	{
		//values are near zero outside of accuracy range and of the different signs
		
		PreciseDouble pd1 = this.preciseDoubleFactory.create(0.000001);
		PreciseDouble pd2 = this.preciseDoubleFactory.create(-0.000001);
		
		assertEquals(0, pd1.decimalPart(), "failed test condition");
		assertEquals(0, pd1.wholePart(), "failed test condition");
		assertEquals(0, pd2.decimalPart(), "failed test condition");
		assertEquals(0, pd2.wholePart(), "failed test condition");
		
		
		assertEquals(pd1, pd2, "Unexpected unequal values.");
	}

	@Test
	void testCompareTo() 
	{
		assertEquals(0, this.preciseDoubleFactory.create(123.123446).compareTo(this.preciseDoubleFactory.create(123.1234)));
		assertEquals(1, this.preciseDoubleFactory.create(123.123456).compareTo(this.preciseDoubleFactory.create(123.1234)));
		assertEquals(-1, this.preciseDoubleFactory.create(123).compareTo(this.preciseDoubleFactory.create(123.1234)));
		assertEquals(1, this.preciseDoubleFactory.create(123.1234).compareTo(this.preciseDoubleFactory.create(123)));
		
		assertEquals(1, this.preciseDoubleFactory.create(124.1234).compareTo(this.preciseDoubleFactory.create(123.5670)));
		assertEquals(-1, this.preciseDoubleFactory.create(123.5).compareTo(this.preciseDoubleFactory.create(124.1)));
		
		
		assertEquals(1, this.preciseDoubleFactory.create(0.1).compareTo(this.preciseDoubleFactory.create(-0.1)));
		assertEquals(-1, this.preciseDoubleFactory.create(-0.1).compareTo(this.preciseDoubleFactory.create(0.1)));
		
		assertEquals(0, this.preciseDoubleFactory.create(0.000001).compareTo(this.preciseDoubleFactory.create(-0.000001)));
	}

	@Test
	void testHashCode()
	{

		int h1 = this.preciseDoubleFactory.create(123.123).hashCode();
		int h2 = this.preciseDoubleFactory.create(0.123).hashCode();
		int h3 = this.preciseDoubleFactory.create(123).hashCode();
		int h4 = this.preciseDoubleFactory.create(0).hashCode();
		int h5 = this.preciseDoubleFactory.create(123.12309).hashCode();
		int h6 = this.preciseDoubleFactory.create(123.12304).hashCode();

		assertNotEquals(h1, h2);
		assertNotEquals(h1, h3);
		assertNotEquals(h1, h4);
		assertNotEquals(h2, h3);
		assertNotEquals(h2, h4);
		assertNotEquals(h3, h4);
		assertNotEquals(h1, h5);
		assertEquals(h1, h6);
	}
}
